from typing import Any

from .local_fs_blob_store import LocalFSBLOBStore
from .metrics.blob_store import (
    metric_abort_multipart_upload_errors,
    metric_abort_multipart_upload_latency,
    metric_abort_multipart_upload_requests,
    metric_complete_multipart_upload_errors,
    metric_complete_multipart_upload_latency,
    metric_complete_multipart_upload_requests,
    metric_create_multipart_upload_errors,
    metric_create_multipart_upload_latency,
    metric_create_multipart_upload_requests,
    metric_get_blob_errors,
    metric_get_blob_latency,
    metric_get_blob_requests,
    metric_presign_uri_errors,
    metric_presign_uri_latency,
    metric_presign_uri_requests,
    metric_upload_blob_errors,
    metric_upload_blob_latency,
    metric_upload_blob_requests,
)
from .s3_blob_store import S3BLOBStore


class BLOBStore:
    """Dispatches generic BLOB store calls to their real backends."""

    def __init__(self):
        self._local: LocalFSBLOBStore = LocalFSBLOBStore()
        self._s3: S3BLOBStore = S3BLOBStore()

    async def get(self, uri: str, logger: Any) -> bytes:
        """Returns binary value stored in BLOB with the supplied URI.

        Raises Exception on error. Raises KeyError if the BLOB doesn't exist.
        """
        with (
            metric_get_blob_errors.count_exceptions(),
            metric_get_blob_latency.time(),
        ):
            metric_get_blob_requests.inc()
            if _is_file_uri(uri):
                return await self._local.get(uri, logger)
            else:
                return await self._s3.get(uri, logger)

    async def presign_get_uri(self, uri: str, expires_in_sec: int, logger: Any) -> str:
        """Returns a presigned URI for getting the BLOB with the supplied URI.

        The URI allows to read any byte range in the BLOB."""
        with (
            metric_presign_uri_errors.count_exceptions(),
            metric_presign_uri_latency.time(),
        ):
            metric_presign_uri_requests.inc()
            if _is_file_uri(uri):
                return await self._local.presign_get_uri(uri, expires_in_sec, logger)
            else:
                return await self._s3.presign_get_uri(uri, expires_in_sec, logger)

    async def upload(self, uri: str, value: bytes, logger: Any) -> None:
        """Stores the supplied binary value in a BLOB with the supplied URI.

        Overwrites existing BLOB. Raises Exception on error.
        """
        with (
            metric_upload_blob_errors.count_exceptions(),
            metric_upload_blob_latency.time(),
        ):
            metric_upload_blob_requests.inc()
            if _is_file_uri(uri):
                await self._local.upload(uri, value, logger)
            else:
                await self._s3.upload(uri, value, logger)

    async def create_multipart_upload(self, uri: str, logger: Any) -> str:
        """Creates a multipart upload for BLOB with the supplied URI and returns the upload ID."""
        with (
            metric_create_multipart_upload_errors.count_exceptions(),
            metric_create_multipart_upload_latency.time(),
        ):
            metric_create_multipart_upload_requests.inc()
            if _is_file_uri(uri):
                return await self._local.create_multipart_upload(uri, logger)
            else:
                return await self._s3.create_multipart_upload(uri, logger)

    async def complete_multipart_upload(
        self, uri: str, upload_id: str, parts_etags: list[str], logger: Any
    ) -> None:
        """Completes a multipart upload for BLOB with the supplied URI.

        parts_etags is a list of ETags for the parts that were uploaded.
        The list is ordered by part number starting from 1.
        """
        with (
            metric_complete_multipart_upload_errors.count_exceptions(),
            metric_complete_multipart_upload_latency.time(),
        ):
            metric_complete_multipart_upload_requests.inc()
            if _is_file_uri(uri):
                await self._local.complete_multipart_upload(
                    uri, upload_id, parts_etags, logger
                )
            else:
                await self._s3.complete_multipart_upload(
                    uri, upload_id, parts_etags, logger
                )

    async def abort_multipart_upload(
        self, uri: str, upload_id: str, logger: Any
    ) -> None:
        """Aborts a multipart upload for BLOB with the supplied URI."""
        with (
            metric_abort_multipart_upload_errors.count_exceptions(),
            metric_abort_multipart_upload_latency.time(),
        ):
            metric_abort_multipart_upload_requests.inc()
            if _is_file_uri(uri):
                await self._local.abort_multipart_upload(uri, upload_id, logger)
            else:
                await self._s3.abort_multipart_upload(uri, upload_id, logger)

    async def presign_upload_part_uri(
        self,
        uri: str,
        part_number: int,
        upload_id: str,
        expires_in_sec: int,
        logger: Any,
    ) -> str:
        """Returns a presigned URI for uploading a part in a multipart upload.

        part_number starts from 1."""
        with (
            metric_presign_uri_errors.count_exceptions(),
            metric_presign_uri_latency.time(),
        ):
            metric_presign_uri_requests.inc()
            if _is_file_uri(uri):
                return await self._local.presign_upload_part_uri(
                    uri, part_number, upload_id, expires_in_sec, logger
                )
            else:
                return await self._s3.presign_upload_part_uri(
                    uri, part_number, upload_id, expires_in_sec, logger
                )


def _is_file_uri(uri: str) -> bool:
    return uri.startswith("file://")



from typing import Any, Dict

from tensorlake.function_executor.proto.function_executor_pb2 import (
    BLOB,
    ExecutionPlanUpdates,
    SerializedObjectInsideBLOB,
)

from indexify.proto.executor_api_pb2 import (
    Allocation,
    AllocationFailureReason,
    AllocationOutcomeCode,
    FunctionExecutorTerminationReason,
)


class AllocationMetrics:
    """Metrics for an allocation."""

    def __init__(self, counters: Dict[str, int], timers: Dict[str, float]):
        self.counters = counters
        self.timers = timers


class AllocationOutput:
    """Result of running an allocation."""

    def __init__(
        self,
        allocation: Allocation,
        outcome_code: AllocationOutcomeCode,
        failure_reason: AllocationFailureReason | None = None,
        output_value: SerializedObjectInsideBLOB | None = None,
        output_execution_plan_updates: ExecutionPlanUpdates | None = None,
        uploaded_function_outputs_blob: BLOB | None = None,
        request_error_output: SerializedObjectInsideBLOB | None = None,
        uploaded_request_error_blob: BLOB | None = None,
        metrics: AllocationMetrics | None = None,
        execution_start_time: float | None = None,
        execution_end_time: float | None = None,
    ):
        self.allocation = allocation
        self.outcome_code = outcome_code
        self.failure_reason = failure_reason
        self.output_value = output_value
        self.output_execution_plan_updates = output_execution_plan_updates
        self.uploaded_function_outputs_blob = uploaded_function_outputs_blob
        self.request_error_output = request_error_output
        self.uploaded_request_error_blob = uploaded_request_error_blob
        self.metrics = metrics
        self.execution_start_time = execution_start_time
        self.execution_end_time = execution_end_time

    @classmethod
    def internal_error(
        cls,
        allocation: Allocation,
        execution_start_time: float | None,
        execution_end_time: float | None,
    ) -> "AllocationOutput":
        """Creates a AllocationOutput for an internal error."""
        return AllocationOutput(
            allocation=allocation,
            outcome_code=AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE,
            failure_reason=AllocationFailureReason.ALLOCATION_FAILURE_REASON_INTERNAL_ERROR,
            execution_start_time=execution_start_time,
            execution_end_time=execution_end_time,
        )

    @classmethod
    def function_timeout(
        cls,
        allocation: Allocation,
        execution_start_time: float | None,
        execution_end_time: float | None,
    ) -> "AllocationOutput":
        """Creates a AllocationOutput for a function timeout error."""
        return AllocationOutput(
            allocation=allocation,
            outcome_code=AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE,
            failure_reason=AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_TIMEOUT,
            execution_start_time=execution_start_time,
            execution_end_time=execution_end_time,
        )

    @classmethod
    def function_executor_unresponsive(
        cls,
        allocation: Allocation,
        execution_start_time: float | None,
        execution_end_time: float | None,
    ) -> "AllocationOutput":
        """Creates a AllocationOutput for an unresponsive FE aka grey failure."""
        # When FE is unresponsive we don't know exact cause of the failure.
        return AllocationOutput(
            allocation=allocation,
            outcome_code=AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE,
            # Treat the grey failure as a function error and thus charge the customer.
            # This is to prevent service abuse by intentionally misbehaving functions.
            failure_reason=AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_ERROR,
            execution_start_time=execution_start_time,
            execution_end_time=execution_end_time,
        )

    @classmethod
    def allocation_cancelled(
        cls,
        allocation: Allocation,
        execution_start_time: float | None,
        execution_end_time: float | None,
    ) -> "AllocationOutput":
        """Creates a AllocationOutput for the case when allocation didn't finish because its allocation was removed by Server."""
        return AllocationOutput(
            allocation=allocation,
            outcome_code=AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE,
            failure_reason=AllocationFailureReason.ALLOCATION_FAILURE_REASON_ALLOCATION_CANCELLED,
            execution_start_time=execution_start_time,
            execution_end_time=execution_end_time,
        )

    @classmethod
    def function_executor_terminated(
        cls,
        allocation: Allocation,
    ) -> "AllocationOutput":
        """Creates a AllocationOutput for the case when allocation didn't run because its FE terminated."""
        return AllocationOutput(
            allocation=allocation,
            outcome_code=AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE,
            failure_reason=AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED,
        )

    @classmethod
    def function_executor_startup_failed(
        cls,
        allocation: Allocation,
        fe_termination_reason: FunctionExecutorTerminationReason,
        logger: Any,
    ) -> "AllocationOutput":
        """Creates AllocationOutput for the case when we fail an allocation that didn't run because its FE startup failed."""
        return AllocationOutput(
            allocation=allocation,
            outcome_code=AllocationOutcomeCode.ALLOCATION_OUTCOME_CODE_FAILURE,
            failure_reason=_fe_startup_failure_reason_to_alloc_failure_reason(
                fe_termination_reason, logger
            ),
        )


def _fe_startup_failure_reason_to_alloc_failure_reason(
    fe_termination_reason: FunctionExecutorTerminationReason, logger: Any
) -> AllocationFailureReason:
    # Only need to check FE termination reasons happening on FE startup.
    if (
        fe_termination_reason
        == FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR
    ):
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_ERROR
    elif (
        fe_termination_reason
        == FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_TIMEOUT
    ):
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_TIMEOUT
    elif (
        fe_termination_reason
        == FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_INTERNAL_ERROR
    ):
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_INTERNAL_ERROR
    elif (
        fe_termination_reason
        == FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED
    ):
        # This fe termination reason is used when FE gets deleted by Server from desired state while it's starting up.
        return (
            AllocationFailureReason.ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED
        )
    else:
        logger.error(
            "unexpected function executor startup failure reason",
            fe_termination_reason=FunctionExecutorTerminationReason.Name(
                fe_termination_reason
            ),
        )
        return AllocationFailureReason.ALLOCATION_FAILURE_REASON_INTERNAL_ERROR



from aiohttp import web


class Handler:
    """Abstract base class for all request handlers."""

    async def handle(self, request: web.Request) -> web.Response:
        raise NotImplementedError("Subclasses must implement this method.")



from typing import Any, List

from .nvidia_gpu import NvidiaGPUInfo, fetch_nvidia_gpu_infos, nvidia_gpus_are_available


class NvidiaGPUAllocator:
    """NvidiaGPUAllocator is a class that manages the allocation and deallocation of GPUs."""

    def __init__(self, logger: Any):
        gpu_infos: List[NvidiaGPUInfo] = []

        if nvidia_gpus_are_available():
            gpu_infos = fetch_nvidia_gpu_infos(logger)
            logger.bind(module=__name__).info(
                "Fetched information about NVIDIA GPUs:", info=gpu_infos
            )

        self._all_gpus: List[NvidiaGPUInfo] = gpu_infos
        self._free_gpus: List[NvidiaGPUInfo] = list(gpu_infos)

    def allocate(self, count: int, logger: Any) -> List[NvidiaGPUInfo]:
        """
        Allocates a specified number of GPUs.

        Args:
            count (int): The number of GPUs to allocate.

        Returns:
            List[NvidiaGPUInfo]: A list of allocated GPUs. The list is empty if count is 0.

        Raises:
            ValueError: If the requested number of GPUs exceeds free GPUs.
            Exception: If an error occurs during allocation.
        """
        if count > len(self._free_gpus):
            raise ValueError(
                f"Not enough free GPUs available, requested={count}, available={len(self._free_gpus)}"
            )

        allocated_gpus: List[NvidiaGPUInfo] = []
        for _ in range(count):
            allocated_gpus.append(self._free_gpus.pop())

        if len(allocated_gpus) > 0:
            logger.bind(module=__name__).info("allocated GPUs:", gpus=allocated_gpus)

        return allocated_gpus

    def deallocate(self, gpus: List[NvidiaGPUInfo], logger: Any) -> None:
        self._free_gpus.extend(gpus)

        if len(gpus) > 0:
            logger.bind(module=__name__).info("deallocated GPUs:", gpus=gpus)

    def list_all(self) -> List[NvidiaGPUInfo]:
        return list(self._all_gpus)  # Return a copy to avoid external modification

    def list_free(self) -> List[NvidiaGPUInfo]:
        return list(self._free_gpus)  # Return a copy to avoid external modification



import asyncio
from typing import Any

import grpc

from .client_configuration import GRPC_CHANNEL_OPTIONS
from .function_executor_server import FunctionExecutorServer


class SubprocessFunctionExecutorServer(FunctionExecutorServer):
    """A FunctionExecutorServer that runs in a child process."""

    def __init__(
        self,
        process: asyncio.subprocess.Process,
        port: int,
        address: str,
    ):
        self._proc = process
        self._port = port
        self._address = address

    async def create_channel(self, logger: Any) -> grpc.aio.Channel:
        return grpc.aio.insecure_channel(self._address, options=GRPC_CHANNEL_OPTIONS)



import os
from typing import Any

import httpx
from pydantic import BaseModel

from .metrics.executor import (
    metric_executor_event_push_errors,
    metric_executor_events_pushed,
)

ENVIRONMENT_EVENT_COLLECTOR_URL = os.environ.get("TENSORLAKE_EVENT_COLLECTOR_URL")


class Resource(BaseModel):
    namespace: str
    application: str
    application_version: str
    executor_id: str
    fn_executor_id: str
    fn: str


class EventCollector:
    def __init__(
        self, logger: Any, collector_url: str | None = ENVIRONMENT_EVENT_COLLECTOR_URL
    ):
        self._logger: Any = logger.bind(module=__name__)
        self._collector_url: str = collector_url
        self._client = httpx.Client()

    def push_event(self, resource: Resource, event: dict[str, Any]) -> None:
        """
        Pushes the given event to a log collector.
        If the collector_url is not set, the event is ignored.

        This function does not capture any exceptions that may occur during the HTTP request
        because it's designed to be embedded into the executor.
        The executor needs to handle HTTP errors and collect metrics.
        """
        if self._collector_url:
            try:
                metric_executor_events_pushed.inc()

                body = resource.model_dump()
                body["event"] = event

                response = self._client.post(self._collector_url, json=body)
                _ = response.raise_for_status()
            except Exception as e:
                metric_executor_event_push_errors.inc()
                self._logger.error("Failed to push event to collector", exc_info=e)



import prometheus_client

from indexify.executor.monitoring.metrics import latency_metric_for_fast_operation

# Allocation finalization metrics.
metric_allocation_finalizations: prometheus_client.Counter = prometheus_client.Counter(
    "allocation_finalizations",
    "Number of allocation finalizations",
)
metric_allocation_finalization_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "allocation_finalization_errors",
        "Number of allocation finalization errors",
    )
)
metric_allocations_finalizing: prometheus_client.Gauge = prometheus_client.Gauge(
    "allocations_finalizing",
    "Number of allocations currently finalizing",
)
metric_allocation_finalization_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "allocation_finalization", "allocation finalization"
    )
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: indexify/proto/executor_api.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 31, 1, "", "indexify/proto/executor_api.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n!indexify/proto/executor_api.proto\x12\x0f\x65xecutor_api_pb"\xbf\x03\n\x0b\x44\x61taPayload\x12\x10\n\x03uri\x18\x01 \x01(\tH\x00\x88\x01\x01\x12;\n\x08\x65ncoding\x18\x02 \x01(\x0e\x32$.executor_api_pb.DataPayloadEncodingH\x01\x88\x01\x01\x12\x1d\n\x10\x65ncoding_version\x18\x03 \x01(\x04H\x02\x88\x01\x01\x12\x19\n\x0c\x63ontent_type\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x1a\n\rmetadata_size\x18\x05 \x01(\x04H\x04\x88\x01\x01\x12\x13\n\x06offset\x18\x06 \x01(\x04H\x05\x88\x01\x01\x12\x11\n\x04size\x18\x07 \x01(\x04H\x06\x88\x01\x01\x12\x18\n\x0bsha256_hash\x18\x08 \x01(\tH\x07\x88\x01\x01\x12$\n\x17source_function_call_id\x18\t \x01(\tH\x08\x88\x01\x01\x12\x0f\n\x02id\x18\n \x01(\tH\t\x88\x01\x01\x42\x06\n\x04_uriB\x0b\n\t_encodingB\x13\n\x11_encoding_versionB\x0f\n\r_content_typeB\x10\n\x0e_metadata_sizeB\t\n\x07_offsetB\x07\n\x05_sizeB\x0e\n\x0c_sha256_hashB\x1a\n\x18_source_function_call_idB\x05\n\x03_id"e\n\x0cGPUResources\x12\x12\n\x05\x63ount\x18\x01 \x01(\rH\x00\x88\x01\x01\x12-\n\x05model\x18\x02 \x01(\x0e\x32\x19.executor_api_pb.GPUModelH\x01\x88\x01\x01\x42\x08\n\x06_countB\x08\n\x06_model"\xc2\x01\n\rHostResources\x12\x16\n\tcpu_count\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x19\n\x0cmemory_bytes\x18\x02 \x01(\x04H\x01\x88\x01\x01\x12\x17\n\ndisk_bytes\x18\x03 \x01(\x04H\x02\x88\x01\x01\x12/\n\x03gpu\x18\x04 \x01(\x0b\x32\x1d.executor_api_pb.GPUResourcesH\x03\x88\x01\x01\x42\x0c\n\n_cpu_countB\x0f\n\r_memory_bytesB\r\n\x0b_disk_bytesB\x06\n\x04_gpu"\xd3\x01\n\x0f\x41llowedFunction\x12\x16\n\tnamespace\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1d\n\x10\x61pplication_name\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1a\n\rfunction_name\x18\x03 \x01(\tH\x02\x88\x01\x01\x12 \n\x13\x61pplication_version\x18\x04 \x01(\tH\x03\x88\x01\x01\x42\x0c\n\n_namespaceB\x13\n\x11_application_nameB\x10\n\x0e_function_nameB\x16\n\x14_application_version"\xd8\x01\n\x19\x46unctionExecutorResources\x12\x1b\n\x0e\x63pu_ms_per_sec\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x19\n\x0cmemory_bytes\x18\x02 \x01(\x04H\x01\x88\x01\x01\x12\x17\n\ndisk_bytes\x18\x03 \x01(\x04H\x02\x88\x01\x01\x12/\n\x03gpu\x18\x04 \x01(\x0b\x32\x1d.executor_api_pb.GPUResourcesH\x03\x88\x01\x01\x42\x11\n\x0f_cpu_ms_per_secB\x0f\n\r_memory_bytesB\r\n\x0b_disk_bytesB\x06\n\x04_gpu"\xcf\x01\n\x0b\x46unctionRef\x12\x16\n\tnamespace\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1d\n\x10\x61pplication_name\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1a\n\rfunction_name\x18\x03 \x01(\tH\x02\x88\x01\x01\x12 \n\x13\x61pplication_version\x18\x04 \x01(\tH\x03\x88\x01\x01\x42\x0c\n\n_namespaceB\x13\n\x11_application_nameB\x10\n\x0e_function_nameB\x16\n\x14_application_version"\xa3\x04\n\x1b\x46unctionExecutorDescription\x12\x0f\n\x02id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x33\n\x08\x66unction\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12\x14\n\x0csecret_names\x18\x03 \x03(\t\x12&\n\x19initialization_timeout_ms\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x36\n\x0b\x61pplication\x18\x05 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x03\x88\x01\x01\x12\x42\n\tresources\x18\x06 \x01(\x0b\x32*.executor_api_pb.FunctionExecutorResourcesH\x04\x88\x01\x01\x12&\n\x19output_payload_uri_prefix\x18\x07 \x01(\tH\x05\x88\x01\x01\x12\x1c\n\x0fmax_concurrency\x18\x08 \x01(\rH\x06\x88\x01\x01\x12"\n\x15\x61llocation_timeout_ms\x18\t \x01(\rH\x07\x88\x01\x01\x42\x05\n\x03_idB\x0b\n\t_functionB\x1c\n\x1a_initialization_timeout_msB\x0e\n\x0c_applicationB\x0c\n\n_resourcesB\x1c\n\x1a_output_payload_uri_prefixB\x12\n\x10_max_concurrencyB\x18\n\x16_allocation_timeout_ms"\xcf\x02\n\x15\x46unctionExecutorState\x12\x46\n\x0b\x64\x65scription\x18\x01 \x01(\x0b\x32,.executor_api_pb.FunctionExecutorDescriptionH\x00\x88\x01\x01\x12<\n\x06status\x18\x02 \x01(\x0e\x32\'.executor_api_pb.FunctionExecutorStatusH\x01\x88\x01\x01\x12S\n\x12termination_reason\x18\x03 \x01(\x0e\x32\x32.executor_api_pb.FunctionExecutorTerminationReasonH\x02\x88\x01\x01\x12)\n!allocation_ids_caused_termination\x18\x04 \x03(\tB\x0e\n\x0c_descriptionB\t\n\x07_statusB\x15\n\x13_termination_reason"\xce\x05\n\rExecutorState\x12\x18\n\x0b\x65xecutor_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x15\n\x08hostname\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x14\n\x07version\x18\x05 \x01(\tH\x02\x88\x01\x01\x12\x34\n\x06status\x18\x06 \x01(\x0e\x32\x1f.executor_api_pb.ExecutorStatusH\x03\x88\x01\x01\x12<\n\x0ftotal_resources\x18\r \x01(\x0b\x32\x1e.executor_api_pb.HostResourcesH\x04\x88\x01\x01\x12N\n!total_function_executor_resources\x18\x07 \x01(\x0b\x32\x1e.executor_api_pb.HostResourcesH\x05\x88\x01\x01\x12;\n\x11\x61llowed_functions\x18\x08 \x03(\x0b\x32 .executor_api_pb.AllowedFunction\x12H\n\x18\x66unction_executor_states\x18\t \x03(\x0b\x32&.executor_api_pb.FunctionExecutorState\x12:\n\x06labels\x18\n \x03(\x0b\x32*.executor_api_pb.ExecutorState.LabelsEntry\x12\x17\n\nstate_hash\x18\x0b \x01(\tH\x06\x88\x01\x01\x12\x19\n\x0cserver_clock\x18\x0c \x01(\x04H\x07\x88\x01\x01\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0e\n\x0c_executor_idB\x0b\n\t_hostnameB\n\n\x08_versionB\t\n\x07_statusB\x12\n\x10_total_resourcesB$\n"_total_function_executor_resourcesB\r\n\x0b_state_hashB\x0f\n\r_server_clock"y\n\x0e\x45xecutorUpdate\x12\x18\n\x0b\x65xecutor_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12=\n\x12\x61llocation_results\x18\x02 \x03(\x0b\x32!.executor_api_pb.AllocationResultB\x0e\n\x0c_executor_id"\xbf\x01\n\x1aReportExecutorStateRequest\x12;\n\x0e\x65xecutor_state\x18\x01 \x01(\x0b\x32\x1e.executor_api_pb.ExecutorStateH\x00\x88\x01\x01\x12=\n\x0f\x65xecutor_update\x18\x02 \x01(\x0b\x32\x1f.executor_api_pb.ExecutorUpdateH\x01\x88\x01\x01\x42\x11\n\x0f_executor_stateB\x12\n\x10_executor_update"\x1d\n\x1bReportExecutorStateResponse"\x9a\x04\n\nAllocation\x12\x33\n\x08\x66unction\x18\x01 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x00\x88\x01\x01\x12\x1a\n\rallocation_id\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1d\n\x10\x66unction_call_id\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x17\n\nrequest_id\x18\x04 \x01(\tH\x03\x88\x01\x01\x12*\n\x04\x61rgs\x18\x05 \x03(\x0b\x32\x1c.executor_api_pb.DataPayload\x12&\n\x19output_payload_uri_prefix\x18\x06 \x01(\tH\x04\x88\x01\x01\x12-\n request_error_payload_uri_prefix\x18\x07 \x01(\tH\x05\x88\x01\x01\x12!\n\x14\x66unction_executor_id\x18\x08 \x01(\tH\x06\x88\x01\x01\x12#\n\x16\x66unction_call_metadata\x18\t \x01(\x0cH\x07\x88\x01\x01\x42\x0b\n\t_functionB\x10\n\x0e_allocation_idB\x13\n\x11_function_call_idB\r\n\x0b_request_idB\x1c\n\x1a_output_payload_uri_prefixB#\n!_request_error_payload_uri_prefixB\x17\n\x15_function_executor_idB\x19\n\x17_function_call_metadata"K\n\x1fGetDesiredExecutorStatesRequest\x12\x18\n\x0b\x65xecutor_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x0e\n\x0c_executor_id"\xb0\x01\n\x14\x44\x65siredExecutorState\x12H\n\x12\x66unction_executors\x18\x01 \x03(\x0b\x32,.executor_api_pb.FunctionExecutorDescription\x12\x30\n\x0b\x61llocations\x18\x02 \x03(\x0b\x32\x1b.executor_api_pb.Allocation\x12\x12\n\x05\x63lock\x18\x03 \x01(\x04H\x00\x88\x01\x01\x42\x08\n\x06_clock"\x80\x01\n\x13\x45xecutionPlanUpdate\x12\x36\n\rfunction_call\x18\n \x01(\x0b\x32\x1d.executor_api_pb.FunctionCallH\x00\x12+\n\x06reduce\x18\x0c \x01(\x0b\x32\x19.executor_api_pb.ReduceOpH\x00\x42\x04\n\x02op"\xbe\x01\n\x0c\x46unctionCall\x12\x0f\n\x02id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x31\n\x06target\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12*\n\x04\x61rgs\x18\x03 \x03(\x0b\x32\x1c.executor_api_pb.FunctionArg\x12\x1a\n\rcall_metadata\x18\x05 \x01(\x0cH\x02\x88\x01\x01\x42\x05\n\x03_idB\t\n\x07_targetB\x10\n\x0e_call_metadata"h\n\x0b\x46unctionArg\x12\x1a\n\x10\x66unction_call_id\x18\x01 \x01(\tH\x00\x12\x33\n\x0binline_data\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x00\x42\x08\n\x06source"\xc2\x01\n\x08ReduceOp\x12\x0f\n\x02id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x30\n\ncollection\x18\x02 \x03(\x0b\x32\x1c.executor_api_pb.FunctionArg\x12\x32\n\x07reducer\x18\x03 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12\x1a\n\rcall_metadata\x18\x04 \x01(\x0cH\x02\x88\x01\x01\x42\x05\n\x03_idB\n\n\x08_reducerB\x10\n\x0e_call_metadata"\x8b\x01\n\x14\x45xecutionPlanUpdates\x12\x35\n\x07updates\x18\x01 \x03(\x0b\x32$.executor_api_pb.ExecutionPlanUpdate\x12"\n\x15root_function_call_id\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x18\n\x16_root_function_call_id"\x8f\x05\n\x10\x41llocationResult\x12\x33\n\x08\x66unction\x18\x01 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12\x1a\n\rallocation_id\x18\x02 \x01(\tH\x02\x88\x01\x01\x12\x1d\n\x10\x66unction_call_id\x18\x03 \x01(\tH\x03\x88\x01\x01\x12\x17\n\nrequest_id\x18\x04 \x01(\tH\x04\x88\x01\x01\x12\x41\n\x0coutcome_code\x18\x05 \x01(\x0e\x32&.executor_api_pb.AllocationOutcomeCodeH\x05\x88\x01\x01\x12\x45\n\x0e\x66\x61ilure_reason\x18\x06 \x01(\x0e\x32(.executor_api_pb.AllocationFailureReasonH\x06\x88\x01\x01\x12-\n\x05value\x18\x07 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x00\x12\x38\n\x07updates\x18\x08 \x01(\x0b\x32%.executor_api_pb.ExecutionPlanUpdatesH\x00\x12\x38\n\rrequest_error\x18\t \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x07\x88\x01\x01\x12"\n\x15\x65xecution_duration_ms\x18\n \x01(\x04H\x08\x88\x01\x01\x42\x0e\n\x0creturn_valueB\x0b\n\t_functionB\x10\n\x0e_allocation_idB\x13\n\x11_function_call_idB\r\n\x0b_request_idB\x0f\n\r_outcome_codeB\x11\n\x0f_failure_reasonB\x10\n\x0e_request_errorB\x18\n\x16_execution_duration_ms"\xab\x02\n\x13\x46unctionCallRequest\x12\x1e\n\x11parent_request_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x33\n\x08\x66unction\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12,\n\x06inputs\x18\x03 \x03(\x0b\x32\x1c.executor_api_pb.DataPayload\x12\x1a\n\rcall_metadata\x18\x04 \x01(\x0cH\x02\x88\x01\x01\x12$\n\x17source_function_call_id\x18\x05 \x01(\tH\x03\x88\x01\x01\x42\x14\n\x12_parent_request_idB\x0b\n\t_functionB\x10\n\x0e_call_metadataB\x1a\n\x18_source_function_call_id"c\n\x12\x46unctionCallResult\x12.\n\x06output\x18\x01 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x00\x12\x13\n\texception\x18\x02 \x01(\tH\x00\x42\x08\n\x06result"h\n\x14\x46unctionCallResponse\x12\x10\n\x06update\x18\x01 \x01(\tH\x00\x12\x35\n\x06result\x18\x02 \x01(\x0b\x32#.executor_api_pb.FunctionCallResultH\x00\x42\x07\n\x05\x65vent*\xf0\x01\n\x13\x44\x61taPayloadEncoding\x12!\n\x1d\x44\x41TA_PAYLOAD_ENCODING_UNKNOWN\x10\x00\x12#\n\x1f\x44\x41TA_PAYLOAD_ENCODING_UTF8_JSON\x10\x01\x12#\n\x1f\x44\x41TA_PAYLOAD_ENCODING_UTF8_TEXT\x10\x02\x12\'\n#DATA_PAYLOAD_ENCODING_BINARY_PICKLE\x10\x03\x12$\n DATA_PAYLOAD_ENCODING_BINARY_ZIP\x10\x04\x12\x1d\n\x19\x44\x41TA_PAYLOAD_ENCODING_RAW\x10\x05*\xd6\x01\n\x08GPUModel\x12\x15\n\x11GPU_MODEL_UNKNOWN\x10\x00\x12\x1e\n\x1aGPU_MODEL_NVIDIA_A100_40GB\x10\x01\x12\x1e\n\x1aGPU_MODEL_NVIDIA_A100_80GB\x10\x02\x12\x1e\n\x1aGPU_MODEL_NVIDIA_H100_80GB\x10\x03\x12\x1d\n\x19GPU_MODEL_NVIDIA_TESLA_T4\x10\x04\x12\x1a\n\x16GPU_MODEL_NVIDIA_A6000\x10\x05\x12\x18\n\x14GPU_MODEL_NVIDIA_A10\x10\x06*\xb3\x01\n\x16\x46unctionExecutorStatus\x12$\n FUNCTION_EXECUTOR_STATUS_UNKNOWN\x10\x00\x12$\n FUNCTION_EXECUTOR_STATUS_PENDING\x10\x01\x12$\n FUNCTION_EXECUTOR_STATUS_RUNNING\x10\x02\x12\'\n#FUNCTION_EXECUTOR_STATUS_TERMINATED\x10\x03*\x94\x04\n!FunctionExecutorTerminationReason\x12\x30\n,FUNCTION_EXECUTOR_TERMINATION_REASON_UNKNOWN\x10\x00\x12\x46\nBFUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_INTERNAL_ERROR\x10\x01\x12\x46\nBFUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR\x10\x02\x12H\nDFUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_TIMEOUT\x10\x03\x12\x32\n.FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY\x10\x0c\x12\x37\n3FUNCTION_EXECUTOR_TERMINATION_REASON_INTERNAL_ERROR\x10\r\x12\x39\n5FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_TIMEOUT\x10\x0e\x12;\n7FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED\x10\x0f*\xa5\x01\n\x0e\x45xecutorStatus\x12\x1b\n\x17\x45XECUTOR_STATUS_UNKNOWN\x10\x00\x12\x1f\n\x1b\x45XECUTOR_STATUS_STARTING_UP\x10\x01\x12\x1b\n\x17\x45XECUTOR_STATUS_RUNNING\x10\x02\x12\x1b\n\x17\x45XECUTOR_STATUS_DRAINED\x10\x03\x12\x1b\n\x17\x45XECUTOR_STATUS_STOPPED\x10\x04*\x86\x01\n\x15\x41llocationOutcomeCode\x12#\n\x1f\x41LLOCATION_OUTCOME_CODE_UNKNOWN\x10\x00\x12#\n\x1f\x41LLOCATION_OUTCOME_CODE_SUCCESS\x10\x01\x12#\n\x1f\x41LLOCATION_OUTCOME_CODE_FAILURE\x10\x02*\xe9\x02\n\x17\x41llocationFailureReason\x12%\n!ALLOCATION_FAILURE_REASON_UNKNOWN\x10\x00\x12,\n(ALLOCATION_FAILURE_REASON_INTERNAL_ERROR\x10\x01\x12,\n(ALLOCATION_FAILURE_REASON_FUNCTION_ERROR\x10\x02\x12.\n*ALLOCATION_FAILURE_REASON_FUNCTION_TIMEOUT\x10\x03\x12+\n\'ALLOCATION_FAILURE_REASON_REQUEST_ERROR\x10\x04\x12\x32\n.ALLOCATION_FAILURE_REASON_ALLOCATION_CANCELLED\x10\x05\x12:\n6ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED\x10\x06\x32\xe3\x02\n\x0b\x45xecutorAPI\x12t\n\x15report_executor_state\x12+.executor_api_pb.ReportExecutorStateRequest\x1a,.executor_api_pb.ReportExecutorStateResponse"\x00\x12z\n\x1bget_desired_executor_states\x12\x30.executor_api_pb.GetDesiredExecutorStatesRequest\x1a%.executor_api_pb.DesiredExecutorState"\x00\x30\x01\x12\x62\n\x0finvoke_function\x12$.executor_api_pb.FunctionCallRequest\x1a%.executor_api_pb.FunctionCallResponse"\x00\x30\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "indexify.proto.executor_api_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_EXECUTORSTATE_LABELSENTRY"]._loaded_options = None
    _globals["_EXECUTORSTATE_LABELSENTRY"]._serialized_options = b"8\001"
    _globals["_DATAPAYLOADENCODING"]._serialized_start = 6138
    _globals["_DATAPAYLOADENCODING"]._serialized_end = 6378
    _globals["_GPUMODEL"]._serialized_start = 6381
    _globals["_GPUMODEL"]._serialized_end = 6595
    _globals["_FUNCTIONEXECUTORSTATUS"]._serialized_start = 6598
    _globals["_FUNCTIONEXECUTORSTATUS"]._serialized_end = 6777
    _globals["_FUNCTIONEXECUTORTERMINATIONREASON"]._serialized_start = 6780
    _globals["_FUNCTIONEXECUTORTERMINATIONREASON"]._serialized_end = 7312
    _globals["_EXECUTORSTATUS"]._serialized_start = 7315
    _globals["_EXECUTORSTATUS"]._serialized_end = 7480
    _globals["_ALLOCATIONOUTCOMECODE"]._serialized_start = 7483
    _globals["_ALLOCATIONOUTCOMECODE"]._serialized_end = 7617
    _globals["_ALLOCATIONFAILUREREASON"]._serialized_start = 7620
    _globals["_ALLOCATIONFAILUREREASON"]._serialized_end = 7981
    _globals["_DATAPAYLOAD"]._serialized_start = 55
    _globals["_DATAPAYLOAD"]._serialized_end = 502
    _globals["_GPURESOURCES"]._serialized_start = 504
    _globals["_GPURESOURCES"]._serialized_end = 605
    _globals["_HOSTRESOURCES"]._serialized_start = 608
    _globals["_HOSTRESOURCES"]._serialized_end = 802
    _globals["_ALLOWEDFUNCTION"]._serialized_start = 805
    _globals["_ALLOWEDFUNCTION"]._serialized_end = 1016
    _globals["_FUNCTIONEXECUTORRESOURCES"]._serialized_start = 1019
    _globals["_FUNCTIONEXECUTORRESOURCES"]._serialized_end = 1235
    _globals["_FUNCTIONREF"]._serialized_start = 1238
    _globals["_FUNCTIONREF"]._serialized_end = 1445
    _globals["_FUNCTIONEXECUTORDESCRIPTION"]._serialized_start = 1448
    _globals["_FUNCTIONEXECUTORDESCRIPTION"]._serialized_end = 1995
    _globals["_FUNCTIONEXECUTORSTATE"]._serialized_start = 1998
    _globals["_FUNCTIONEXECUTORSTATE"]._serialized_end = 2333
    _globals["_EXECUTORSTATE"]._serialized_start = 2336
    _globals["_EXECUTORSTATE"]._serialized_end = 3054
    _globals["_EXECUTORSTATE_LABELSENTRY"]._serialized_start = 2867
    _globals["_EXECUTORSTATE_LABELSENTRY"]._serialized_end = 2912
    _globals["_EXECUTORUPDATE"]._serialized_start = 3056
    _globals["_EXECUTORUPDATE"]._serialized_end = 3177
    _globals["_REPORTEXECUTORSTATEREQUEST"]._serialized_start = 3180
    _globals["_REPORTEXECUTORSTATEREQUEST"]._serialized_end = 3371
    _globals["_REPORTEXECUTORSTATERESPONSE"]._serialized_start = 3373
    _globals["_REPORTEXECUTORSTATERESPONSE"]._serialized_end = 3402
    _globals["_ALLOCATION"]._serialized_start = 3405
    _globals["_ALLOCATION"]._serialized_end = 3943
    _globals["_GETDESIREDEXECUTORSTATESREQUEST"]._serialized_start = 3945
    _globals["_GETDESIREDEXECUTORSTATESREQUEST"]._serialized_end = 4020
    _globals["_DESIREDEXECUTORSTATE"]._serialized_start = 4023
    _globals["_DESIREDEXECUTORSTATE"]._serialized_end = 4199
    _globals["_EXECUTIONPLANUPDATE"]._serialized_start = 4202
    _globals["_EXECUTIONPLANUPDATE"]._serialized_end = 4330
    _globals["_FUNCTIONCALL"]._serialized_start = 4333
    _globals["_FUNCTIONCALL"]._serialized_end = 4523
    _globals["_FUNCTIONARG"]._serialized_start = 4525
    _globals["_FUNCTIONARG"]._serialized_end = 4629
    _globals["_REDUCEOP"]._serialized_start = 4632
    _globals["_REDUCEOP"]._serialized_end = 4826
    _globals["_EXECUTIONPLANUPDATES"]._serialized_start = 4829
    _globals["_EXECUTIONPLANUPDATES"]._serialized_end = 4968
    _globals["_ALLOCATIONRESULT"]._serialized_start = 4971
    _globals["_ALLOCATIONRESULT"]._serialized_end = 5626
    _globals["_FUNCTIONCALLREQUEST"]._serialized_start = 5629
    _globals["_FUNCTIONCALLREQUEST"]._serialized_end = 5928
    _globals["_FUNCTIONCALLRESULT"]._serialized_start = 5930
    _globals["_FUNCTIONCALLRESULT"]._serialized_end = 6029
    _globals["_FUNCTIONCALLRESPONSE"]._serialized_start = 6031
    _globals["_FUNCTIONCALLRESPONSE"]._serialized_end = 6135
    _globals["_EXECUTORAPI"]._serialized_start = 7984
    _globals["_EXECUTORAPI"]._serialized_end = 8339
# @@protoc_insertion_point(module_scope)



from indexify.executor.logging import (
    configure_development_mode_logging,
    configure_logging_early,
    configure_production_mode_logging,
)

configure_logging_early()

import shutil
from importlib.metadata import version
from pathlib import Path
from socket import gethostname
from typing import Dict, List

import click
import nanoid
import prometheus_client
import structlog

from indexify.executor.blob_store.blob_store import BLOBStore
from indexify.executor.executor import Executor
from indexify.executor.function_executor.server.subprocess_function_executor_server_factory import (
    SubprocessFunctionExecutorServerFactory,
)
from indexify.executor.host_resources.host_resources import HostResourcesProvider
from indexify.executor.host_resources.nvidia_gpu_allocator import NvidiaGPUAllocator
from indexify.executor.monitoring.health_checker.generic_health_checker import (
    GenericHealthChecker,
)


@click.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    short_help="Runs Executor that connects to the Indexify server and starts running its tasks",
)
@click.option(
    "--server-addr",
    "server_address",
    default="localhost:8900",
    help="Address of Indexify HTTP Server to connect to",
)
@click.option(
    "--grpc-server-addr",
    "grpc_server_address",
    default="localhost:8901",
    help="Address of Indexify gRPC Server to connect to",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose logging",
)
@click.option(
    "-vv",
    "--very-verbose",
    is_flag=True,
    default=False,
    help="Very verbose logging",
)
@click.option(
    "-f",
    "--function",
    "function_uris",
    default=[],
    multiple=True,
    help="Functions that the executor will run "
    "specified as <namespace>:<application>:<function>:<version>"
    "version is optional, not specifying it will make the server send any version"
    "of the function. Any number of --function arguments can be passed.",
)
@click.option(
    "--config-path",
    type=click.Path(exists=True, dir_okay=False, file_okay=True, readable=True),
    default=None,
    help="Path to the TLS configuration file",
)
@click.option(
    "--executor-cache-path",
    type=click.Path(exists=False, dir_okay=True, readable=True, writable=True),
    default="~/.indexify/executor_cache",
    help="Path to the executor cache directory",
)
@click.option(
    "--monitoring-server-host",
    default="localhost",
    help="IP address or hostname where to run Executor Monitoring server",
)
@click.option(
    "--monitoring-server-port",
    default=7000,
    type=int,
    help="Port where to run Executor Monitoring server",
)
@click.option(
    "-l",
    "--label",
    "labels",
    default=[],
    multiple=True,
    help="Executor key-value label to be sent to the Server. "
    "Specified as <key>=<value>",
)
@click.pass_context
def executor(
    ctx: click.Context,
    server_address: str,
    grpc_server_address: str,
    verbose: bool,
    very_verbose: bool,
    function_uris: List[str],
    config_path: str | None,
    executor_cache_path: str,
    monitoring_server_host: str,
    monitoring_server_port: int,
    labels: List[str],
):
    if verbose or very_verbose:
        configure_development_mode_logging(compact_tracebacks=not very_verbose)
    else:
        configure_production_mode_logging()

    kv_labels: Dict[str, str] = {}
    for label in labels:
        key, value = label.split("=")
        kv_labels[key] = value

    executor_id: str = nanoid.generate()
    executor_version = version("indexify")
    logger = structlog.get_logger(module=__name__, executor_id=executor_id)

    logger.info(
        "starting executor",
        hostname=gethostname(),
        server_address=server_address,
        grpc_server_address=grpc_server_address,
        config_path=config_path,
        executor_version=executor_version,
        labels=kv_labels,
        executor_cache_path=executor_cache_path,
        functions=function_uris,
        verbose=verbose,
        very_verbose=very_verbose,
        monitoring_server_host=monitoring_server_host,
        monitoring_server_port=monitoring_server_port,
    )
    if ctx.args:
        logger.warning(
            "Unknown arguments passed to the executor",
            unknown_args=ctx.args,
        )
    if len(function_uris) == 0:
        logger.warning(
            "No --function arguments were passed. Executor will run all functions. This scenario is only supported for testing purposes.",
        )

    executor_cache_path: Path = Path(executor_cache_path).expanduser().absolute()
    if executor_cache_path.exists():
        shutil.rmtree(str(executor_cache_path))
    executor_cache_path.mkdir(parents=True, exist_ok=True)

    host_resources_provider: HostResourcesProvider = HostResourcesProvider(
        gpu_allocator=NvidiaGPUAllocator(logger),
        # Assuming a simple setup in OSS where Executor container has a single file system
        # used by all Function Executors and all the container resources are available to all Function Executors.
        function_executors_ephimeral_disks_path="/",
        host_overhead_cpus=0,
        host_overhead_memory_gb=0,
        host_overhead_function_executors_ephimeral_disks_gb=0,
    )

    prometheus_client.Info("cli", "CLI information").info(
        {
            "package": "indexify",
        }
    )

    Executor(
        id=executor_id,
        version=executor_version,
        labels=kv_labels,
        health_checker=GenericHealthChecker(),
        cache_path=executor_cache_path,
        function_uris=function_uris,
        function_executor_server_factory=SubprocessFunctionExecutorServerFactory(
            verbose_logs=verbose or very_verbose
        ),
        server_addr=server_address,
        grpc_server_addr=grpc_server_address,
        config_path=config_path,
        monitoring_server_host=monitoring_server_host,
        monitoring_server_port=monitoring_server_port,
        blob_store=BLOBStore(),
        host_resources_provider=host_resources_provider,
    ).run()



import asyncio
import time
from typing import Any, Dict, Optional

import grpc.aio
import yaml

from .metrics.channel_manager import (
    metric_grpc_server_channel_creation_latency,
    metric_grpc_server_channel_creation_retries,
    metric_grpc_server_channel_creations,
)

_RETRY_INTERVAL_SEC = 5


class ChannelManager:
    def __init__(
        self,
        server_address: str,
        config_path: Optional[str],
        logger: Any,
    ):
        self._logger: Any = logger.bind(module=__name__, server_address=server_address)
        self._server_address: str = server_address
        self._channel_credentials: Optional[grpc.ChannelCredentials] = None
        # Shared channel used by different Executor components to communicate with Server.
        self._shared_channel_lock = asyncio.Lock()
        self._shared_channel: Optional[grpc.aio.Channel] = None

        self._init_tls(config_path)

    def _init_tls(self, config_path: Optional[str]):
        if config_path is None:
            return

        # The same config file format as in Tensorlake SDK HTTP client, see:
        # https://github.com/tensorlakeai/tensorlake/blob/main/src/tensorlake/utils/http_client.py
        with open(config_path, "r") as config_file:
            config = yaml.safe_load(config_file)

        if not config.get("use_tls", False):
            return

        tls_config: Dict[str, str] = config["tls_config"]
        cert_path: Optional[str] = tls_config.get("cert_path", None)
        key_path: Optional[str] = tls_config.get("key_path", None)
        ca_bundle_path: Optional[str] = tls_config.get("ca_bundle_path", None)

        self._logger = self._logger.bind(
            cert_path=cert_path,
            key_path=key_path,
            ca_bundle_path=ca_bundle_path,
        )
        self._logger.info("TLS is enabled for grpc channels to server")

        private_key: Optional[bytes] = None
        certificate_chain: Optional[bytes] = None
        root_certificates: Optional[bytes] = None

        if cert_path is not None:
            with open(cert_path, "rb") as cert_file:
                certificate_chain = cert_file.read()
        if key_path is not None:
            with open(key_path, "rb") as key_file:
                private_key = key_file.read()
        if ca_bundle_path is not None:
            with open(ca_bundle_path, "rb") as ca_bundle_file:
                root_certificates = ca_bundle_file.read()

        self._channel_credentials = grpc.ssl_channel_credentials(
            root_certificates=root_certificates,
            private_key=private_key,
            certificate_chain=certificate_chain,
        )

    async def destroy(self):
        # Okay to not hold the lock here as we're destroying the server channel forever.
        if self._shared_channel is not None:
            await self._destroy_shared_channel()

    async def fail_shared_channel(self) -> None:
        """Marks the shared channel as unhealthy and creates a new one.

        Doesn't raise any exceptions.
        """
        async with self._shared_channel_lock:
            if self._shared_channel is None:
                self._logger.error(
                    "grpc server channel doesn't exist, can't mark it unhealthy"
                )
                return

            self._logger.info("marking grpc server channel as unhealthy")
            # All the channel users will see it failing cause we destroyed it and call get_channel() again.
            await self._destroy_shared_channel()

    async def get_shared_channel(self) -> grpc.aio.Channel:
        """Returns shared channel to the gRPC server.

        The health of the shared channel is constantly monitored so it's more reliable than using a
        standalone channel created for a particular short term need. Doesn't raise any exceptions.
        """
        # Use the lock to ensure that we only create one channel without race conditions.
        async with self._shared_channel_lock:
            if self._shared_channel is None:
                await self._create_shared_channel()

            return self._shared_channel

    def create_standalone_channel(self) -> grpc.aio.Channel:
        """Creates a new channel to the gRPC server.

        Used for one-off RPCs where we don't need to monitor channel health or retry its creation indefinitely.
        Raises an exception on failure.
        """
        with (
            metric_grpc_server_channel_creation_retries.count_exceptions(),
            metric_grpc_server_channel_creation_latency.time(),
        ):
            metric_grpc_server_channel_creations.inc()
            if self._channel_credentials is None:
                return grpc.aio.insecure_channel(target=self._server_address)
            else:
                return grpc.aio.secure_channel(
                    target=self._server_address,
                    credentials=self._channel_credentials,
                )

    async def _create_shared_channel(self) -> None:
        """Creates new shared channel.

        self._shared_channel_lock must be acquired before calling this method.
        Never raises any exceptions.
        """
        while True:
            try:
                create_channel_start = time.monotonic()
                self._logger.info("creating new grpc channel to server")
                self._shared_channel = self.create_standalone_channel()
                # Ensure the channel tried to connect to not get "channel closed errors" without actually trying to connect.
                self._shared_channel.get_state(try_to_connect=True)
                self._logger.info(
                    "created new grpc channel to server",
                    duration_sec=time.monotonic() - create_channel_start,
                )
                break
            except Exception as e:
                self._logger.error(
                    f"failed creating grpc channel to server, retrying in {_RETRY_INTERVAL_SEC} seconds",
                    exc_info=e,
                )
                await asyncio.sleep(_RETRY_INTERVAL_SEC)

    async def _destroy_shared_channel(self) -> None:
        """Closes the existing shared channel.

        self._shared_channel_lock must be acquired before calling this method.
        Never raises any exceptions.
        """
        try:
            self._logger.info("closing grpc channel to server")
            await self._shared_channel.close()
            self._logger.info("closed grpc channel to server")
        except Exception as e:
            self._logger.error("failed closing grpc channel to server", exc_info=e)
        self._shared_channel = None



from collections.abc import Iterable as _Iterable
from collections.abc import Mapping as _Mapping
from typing import ClassVar as _ClassVar
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

DESCRIPTOR: _descriptor.FileDescriptor

class DataPayloadEncoding(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DATA_PAYLOAD_ENCODING_UNKNOWN: _ClassVar[DataPayloadEncoding]
    DATA_PAYLOAD_ENCODING_UTF8_JSON: _ClassVar[DataPayloadEncoding]
    DATA_PAYLOAD_ENCODING_UTF8_TEXT: _ClassVar[DataPayloadEncoding]
    DATA_PAYLOAD_ENCODING_BINARY_PICKLE: _ClassVar[DataPayloadEncoding]
    DATA_PAYLOAD_ENCODING_BINARY_ZIP: _ClassVar[DataPayloadEncoding]
    DATA_PAYLOAD_ENCODING_RAW: _ClassVar[DataPayloadEncoding]

class GPUModel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    GPU_MODEL_UNKNOWN: _ClassVar[GPUModel]
    GPU_MODEL_NVIDIA_A100_40GB: _ClassVar[GPUModel]
    GPU_MODEL_NVIDIA_A100_80GB: _ClassVar[GPUModel]
    GPU_MODEL_NVIDIA_H100_80GB: _ClassVar[GPUModel]
    GPU_MODEL_NVIDIA_TESLA_T4: _ClassVar[GPUModel]
    GPU_MODEL_NVIDIA_A6000: _ClassVar[GPUModel]
    GPU_MODEL_NVIDIA_A10: _ClassVar[GPUModel]

class FunctionExecutorStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FUNCTION_EXECUTOR_STATUS_UNKNOWN: _ClassVar[FunctionExecutorStatus]
    FUNCTION_EXECUTOR_STATUS_PENDING: _ClassVar[FunctionExecutorStatus]
    FUNCTION_EXECUTOR_STATUS_RUNNING: _ClassVar[FunctionExecutorStatus]
    FUNCTION_EXECUTOR_STATUS_TERMINATED: _ClassVar[FunctionExecutorStatus]

class FunctionExecutorTerminationReason(
    int, metaclass=_enum_type_wrapper.EnumTypeWrapper
):
    __slots__ = ()
    FUNCTION_EXECUTOR_TERMINATION_REASON_UNKNOWN: _ClassVar[
        FunctionExecutorTerminationReason
    ]
    FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_INTERNAL_ERROR: _ClassVar[
        FunctionExecutorTerminationReason
    ]
    FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR: _ClassVar[
        FunctionExecutorTerminationReason
    ]
    FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_TIMEOUT: _ClassVar[
        FunctionExecutorTerminationReason
    ]
    FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY: _ClassVar[
        FunctionExecutorTerminationReason
    ]
    FUNCTION_EXECUTOR_TERMINATION_REASON_INTERNAL_ERROR: _ClassVar[
        FunctionExecutorTerminationReason
    ]
    FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_TIMEOUT: _ClassVar[
        FunctionExecutorTerminationReason
    ]
    FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED: _ClassVar[
        FunctionExecutorTerminationReason
    ]

class ExecutorStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    EXECUTOR_STATUS_UNKNOWN: _ClassVar[ExecutorStatus]
    EXECUTOR_STATUS_STARTING_UP: _ClassVar[ExecutorStatus]
    EXECUTOR_STATUS_RUNNING: _ClassVar[ExecutorStatus]
    EXECUTOR_STATUS_DRAINED: _ClassVar[ExecutorStatus]
    EXECUTOR_STATUS_STOPPED: _ClassVar[ExecutorStatus]

class AllocationOutcomeCode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ALLOCATION_OUTCOME_CODE_UNKNOWN: _ClassVar[AllocationOutcomeCode]
    ALLOCATION_OUTCOME_CODE_SUCCESS: _ClassVar[AllocationOutcomeCode]
    ALLOCATION_OUTCOME_CODE_FAILURE: _ClassVar[AllocationOutcomeCode]

class AllocationFailureReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ALLOCATION_FAILURE_REASON_UNKNOWN: _ClassVar[AllocationFailureReason]
    ALLOCATION_FAILURE_REASON_INTERNAL_ERROR: _ClassVar[AllocationFailureReason]
    ALLOCATION_FAILURE_REASON_FUNCTION_ERROR: _ClassVar[AllocationFailureReason]
    ALLOCATION_FAILURE_REASON_FUNCTION_TIMEOUT: _ClassVar[AllocationFailureReason]
    ALLOCATION_FAILURE_REASON_REQUEST_ERROR: _ClassVar[AllocationFailureReason]
    ALLOCATION_FAILURE_REASON_ALLOCATION_CANCELLED: _ClassVar[AllocationFailureReason]
    ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED: _ClassVar[
        AllocationFailureReason
    ]

DATA_PAYLOAD_ENCODING_UNKNOWN: DataPayloadEncoding
DATA_PAYLOAD_ENCODING_UTF8_JSON: DataPayloadEncoding
DATA_PAYLOAD_ENCODING_UTF8_TEXT: DataPayloadEncoding
DATA_PAYLOAD_ENCODING_BINARY_PICKLE: DataPayloadEncoding
DATA_PAYLOAD_ENCODING_BINARY_ZIP: DataPayloadEncoding
DATA_PAYLOAD_ENCODING_RAW: DataPayloadEncoding
GPU_MODEL_UNKNOWN: GPUModel
GPU_MODEL_NVIDIA_A100_40GB: GPUModel
GPU_MODEL_NVIDIA_A100_80GB: GPUModel
GPU_MODEL_NVIDIA_H100_80GB: GPUModel
GPU_MODEL_NVIDIA_TESLA_T4: GPUModel
GPU_MODEL_NVIDIA_A6000: GPUModel
GPU_MODEL_NVIDIA_A10: GPUModel
FUNCTION_EXECUTOR_STATUS_UNKNOWN: FunctionExecutorStatus
FUNCTION_EXECUTOR_STATUS_PENDING: FunctionExecutorStatus
FUNCTION_EXECUTOR_STATUS_RUNNING: FunctionExecutorStatus
FUNCTION_EXECUTOR_STATUS_TERMINATED: FunctionExecutorStatus
FUNCTION_EXECUTOR_TERMINATION_REASON_UNKNOWN: FunctionExecutorTerminationReason
FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_INTERNAL_ERROR: (
    FunctionExecutorTerminationReason
)
FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR: (
    FunctionExecutorTerminationReason
)
FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_TIMEOUT: (
    FunctionExecutorTerminationReason
)
FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY: FunctionExecutorTerminationReason
FUNCTION_EXECUTOR_TERMINATION_REASON_INTERNAL_ERROR: FunctionExecutorTerminationReason
FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_TIMEOUT: FunctionExecutorTerminationReason
FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED: (
    FunctionExecutorTerminationReason
)
EXECUTOR_STATUS_UNKNOWN: ExecutorStatus
EXECUTOR_STATUS_STARTING_UP: ExecutorStatus
EXECUTOR_STATUS_RUNNING: ExecutorStatus
EXECUTOR_STATUS_DRAINED: ExecutorStatus
EXECUTOR_STATUS_STOPPED: ExecutorStatus
ALLOCATION_OUTCOME_CODE_UNKNOWN: AllocationOutcomeCode
ALLOCATION_OUTCOME_CODE_SUCCESS: AllocationOutcomeCode
ALLOCATION_OUTCOME_CODE_FAILURE: AllocationOutcomeCode
ALLOCATION_FAILURE_REASON_UNKNOWN: AllocationFailureReason
ALLOCATION_FAILURE_REASON_INTERNAL_ERROR: AllocationFailureReason
ALLOCATION_FAILURE_REASON_FUNCTION_ERROR: AllocationFailureReason
ALLOCATION_FAILURE_REASON_FUNCTION_TIMEOUT: AllocationFailureReason
ALLOCATION_FAILURE_REASON_REQUEST_ERROR: AllocationFailureReason
ALLOCATION_FAILURE_REASON_ALLOCATION_CANCELLED: AllocationFailureReason
ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED: AllocationFailureReason

class DataPayload(_message.Message):
    __slots__ = (
        "uri",
        "encoding",
        "encoding_version",
        "content_type",
        "metadata_size",
        "offset",
        "size",
        "sha256_hash",
        "source_function_call_id",
        "id",
    )
    URI_FIELD_NUMBER: _ClassVar[int]
    ENCODING_FIELD_NUMBER: _ClassVar[int]
    ENCODING_VERSION_FIELD_NUMBER: _ClassVar[int]
    CONTENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    METADATA_SIZE_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    SIZE_FIELD_NUMBER: _ClassVar[int]
    SHA256_HASH_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FUNCTION_CALL_ID_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    uri: str
    encoding: DataPayloadEncoding
    encoding_version: int
    content_type: str
    metadata_size: int
    offset: int
    size: int
    sha256_hash: str
    source_function_call_id: str
    id: str
    def __init__(
        self,
        uri: _Optional[str] = ...,
        encoding: _Optional[_Union[DataPayloadEncoding, str]] = ...,
        encoding_version: _Optional[int] = ...,
        content_type: _Optional[str] = ...,
        metadata_size: _Optional[int] = ...,
        offset: _Optional[int] = ...,
        size: _Optional[int] = ...,
        sha256_hash: _Optional[str] = ...,
        source_function_call_id: _Optional[str] = ...,
        id: _Optional[str] = ...,
    ) -> None: ...

class GPUResources(_message.Message):
    __slots__ = ("count", "model")
    COUNT_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    count: int
    model: GPUModel
    def __init__(
        self, count: _Optional[int] = ..., model: _Optional[_Union[GPUModel, str]] = ...
    ) -> None: ...

class HostResources(_message.Message):
    __slots__ = ("cpu_count", "memory_bytes", "disk_bytes", "gpu")
    CPU_COUNT_FIELD_NUMBER: _ClassVar[int]
    MEMORY_BYTES_FIELD_NUMBER: _ClassVar[int]
    DISK_BYTES_FIELD_NUMBER: _ClassVar[int]
    GPU_FIELD_NUMBER: _ClassVar[int]
    cpu_count: int
    memory_bytes: int
    disk_bytes: int
    gpu: GPUResources
    def __init__(
        self,
        cpu_count: _Optional[int] = ...,
        memory_bytes: _Optional[int] = ...,
        disk_bytes: _Optional[int] = ...,
        gpu: _Optional[_Union[GPUResources, _Mapping]] = ...,
    ) -> None: ...

class AllowedFunction(_message.Message):
    __slots__ = (
        "namespace",
        "application_name",
        "function_name",
        "application_version",
    )
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    APPLICATION_NAME_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_NAME_FIELD_NUMBER: _ClassVar[int]
    APPLICATION_VERSION_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    application_name: str
    function_name: str
    application_version: str
    def __init__(
        self,
        namespace: _Optional[str] = ...,
        application_name: _Optional[str] = ...,
        function_name: _Optional[str] = ...,
        application_version: _Optional[str] = ...,
    ) -> None: ...

class FunctionExecutorResources(_message.Message):
    __slots__ = ("cpu_ms_per_sec", "memory_bytes", "disk_bytes", "gpu")
    CPU_MS_PER_SEC_FIELD_NUMBER: _ClassVar[int]
    MEMORY_BYTES_FIELD_NUMBER: _ClassVar[int]
    DISK_BYTES_FIELD_NUMBER: _ClassVar[int]
    GPU_FIELD_NUMBER: _ClassVar[int]
    cpu_ms_per_sec: int
    memory_bytes: int
    disk_bytes: int
    gpu: GPUResources
    def __init__(
        self,
        cpu_ms_per_sec: _Optional[int] = ...,
        memory_bytes: _Optional[int] = ...,
        disk_bytes: _Optional[int] = ...,
        gpu: _Optional[_Union[GPUResources, _Mapping]] = ...,
    ) -> None: ...

class FunctionRef(_message.Message):
    __slots__ = (
        "namespace",
        "application_name",
        "function_name",
        "application_version",
    )
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    APPLICATION_NAME_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_NAME_FIELD_NUMBER: _ClassVar[int]
    APPLICATION_VERSION_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    application_name: str
    function_name: str
    application_version: str
    def __init__(
        self,
        namespace: _Optional[str] = ...,
        application_name: _Optional[str] = ...,
        function_name: _Optional[str] = ...,
        application_version: _Optional[str] = ...,
    ) -> None: ...

class FunctionExecutorDescription(_message.Message):
    __slots__ = (
        "id",
        "function",
        "secret_names",
        "initialization_timeout_ms",
        "application",
        "resources",
        "output_payload_uri_prefix",
        "max_concurrency",
        "allocation_timeout_ms",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    SECRET_NAMES_FIELD_NUMBER: _ClassVar[int]
    INITIALIZATION_TIMEOUT_MS_FIELD_NUMBER: _ClassVar[int]
    APPLICATION_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_PAYLOAD_URI_PREFIX_FIELD_NUMBER: _ClassVar[int]
    MAX_CONCURRENCY_FIELD_NUMBER: _ClassVar[int]
    ALLOCATION_TIMEOUT_MS_FIELD_NUMBER: _ClassVar[int]
    id: str
    function: FunctionRef
    secret_names: _containers.RepeatedScalarFieldContainer[str]
    initialization_timeout_ms: int
    application: DataPayload
    resources: FunctionExecutorResources
    output_payload_uri_prefix: str
    max_concurrency: int
    allocation_timeout_ms: int
    def __init__(
        self,
        id: _Optional[str] = ...,
        function: _Optional[_Union[FunctionRef, _Mapping]] = ...,
        secret_names: _Optional[_Iterable[str]] = ...,
        initialization_timeout_ms: _Optional[int] = ...,
        application: _Optional[_Union[DataPayload, _Mapping]] = ...,
        resources: _Optional[_Union[FunctionExecutorResources, _Mapping]] = ...,
        output_payload_uri_prefix: _Optional[str] = ...,
        max_concurrency: _Optional[int] = ...,
        allocation_timeout_ms: _Optional[int] = ...,
    ) -> None: ...

class FunctionExecutorState(_message.Message):
    __slots__ = (
        "description",
        "status",
        "termination_reason",
        "allocation_ids_caused_termination",
    )
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TERMINATION_REASON_FIELD_NUMBER: _ClassVar[int]
    ALLOCATION_IDS_CAUSED_TERMINATION_FIELD_NUMBER: _ClassVar[int]
    description: FunctionExecutorDescription
    status: FunctionExecutorStatus
    termination_reason: FunctionExecutorTerminationReason
    allocation_ids_caused_termination: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        description: _Optional[_Union[FunctionExecutorDescription, _Mapping]] = ...,
        status: _Optional[_Union[FunctionExecutorStatus, str]] = ...,
        termination_reason: _Optional[
            _Union[FunctionExecutorTerminationReason, str]
        ] = ...,
        allocation_ids_caused_termination: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class ExecutorState(_message.Message):
    __slots__ = (
        "executor_id",
        "hostname",
        "version",
        "status",
        "total_resources",
        "total_function_executor_resources",
        "allowed_functions",
        "function_executor_states",
        "labels",
        "state_hash",
        "server_clock",
    )

    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    HOSTNAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    TOTAL_FUNCTION_EXECUTOR_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    ALLOWED_FUNCTIONS_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_EXECUTOR_STATES_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    STATE_HASH_FIELD_NUMBER: _ClassVar[int]
    SERVER_CLOCK_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    hostname: str
    version: str
    status: ExecutorStatus
    total_resources: HostResources
    total_function_executor_resources: HostResources
    allowed_functions: _containers.RepeatedCompositeFieldContainer[AllowedFunction]
    function_executor_states: _containers.RepeatedCompositeFieldContainer[
        FunctionExecutorState
    ]
    labels: _containers.ScalarMap[str, str]
    state_hash: str
    server_clock: int
    def __init__(
        self,
        executor_id: _Optional[str] = ...,
        hostname: _Optional[str] = ...,
        version: _Optional[str] = ...,
        status: _Optional[_Union[ExecutorStatus, str]] = ...,
        total_resources: _Optional[_Union[HostResources, _Mapping]] = ...,
        total_function_executor_resources: _Optional[
            _Union[HostResources, _Mapping]
        ] = ...,
        allowed_functions: _Optional[
            _Iterable[_Union[AllowedFunction, _Mapping]]
        ] = ...,
        function_executor_states: _Optional[
            _Iterable[_Union[FunctionExecutorState, _Mapping]]
        ] = ...,
        labels: _Optional[_Mapping[str, str]] = ...,
        state_hash: _Optional[str] = ...,
        server_clock: _Optional[int] = ...,
    ) -> None: ...

class ExecutorUpdate(_message.Message):
    __slots__ = ("executor_id", "allocation_results")
    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    ALLOCATION_RESULTS_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    allocation_results: _containers.RepeatedCompositeFieldContainer[AllocationResult]
    def __init__(
        self,
        executor_id: _Optional[str] = ...,
        allocation_results: _Optional[
            _Iterable[_Union[AllocationResult, _Mapping]]
        ] = ...,
    ) -> None: ...

class ReportExecutorStateRequest(_message.Message):
    __slots__ = ("executor_state", "executor_update")
    EXECUTOR_STATE_FIELD_NUMBER: _ClassVar[int]
    EXECUTOR_UPDATE_FIELD_NUMBER: _ClassVar[int]
    executor_state: ExecutorState
    executor_update: ExecutorUpdate
    def __init__(
        self,
        executor_state: _Optional[_Union[ExecutorState, _Mapping]] = ...,
        executor_update: _Optional[_Union[ExecutorUpdate, _Mapping]] = ...,
    ) -> None: ...

class ReportExecutorStateResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class Allocation(_message.Message):
    __slots__ = (
        "function",
        "allocation_id",
        "function_call_id",
        "request_id",
        "args",
        "output_payload_uri_prefix",
        "request_error_payload_uri_prefix",
        "function_executor_id",
        "function_call_metadata",
    )
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    ALLOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CALL_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_PAYLOAD_URI_PREFIX_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ERROR_PAYLOAD_URI_PREFIX_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CALL_METADATA_FIELD_NUMBER: _ClassVar[int]
    function: FunctionRef
    allocation_id: str
    function_call_id: str
    request_id: str
    args: _containers.RepeatedCompositeFieldContainer[DataPayload]
    output_payload_uri_prefix: str
    request_error_payload_uri_prefix: str
    function_executor_id: str
    function_call_metadata: bytes
    def __init__(
        self,
        function: _Optional[_Union[FunctionRef, _Mapping]] = ...,
        allocation_id: _Optional[str] = ...,
        function_call_id: _Optional[str] = ...,
        request_id: _Optional[str] = ...,
        args: _Optional[_Iterable[_Union[DataPayload, _Mapping]]] = ...,
        output_payload_uri_prefix: _Optional[str] = ...,
        request_error_payload_uri_prefix: _Optional[str] = ...,
        function_executor_id: _Optional[str] = ...,
        function_call_metadata: _Optional[bytes] = ...,
    ) -> None: ...

class GetDesiredExecutorStatesRequest(_message.Message):
    __slots__ = ("executor_id",)
    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    def __init__(self, executor_id: _Optional[str] = ...) -> None: ...

class DesiredExecutorState(_message.Message):
    __slots__ = ("function_executors", "allocations", "clock")
    FUNCTION_EXECUTORS_FIELD_NUMBER: _ClassVar[int]
    ALLOCATIONS_FIELD_NUMBER: _ClassVar[int]
    CLOCK_FIELD_NUMBER: _ClassVar[int]
    function_executors: _containers.RepeatedCompositeFieldContainer[
        FunctionExecutorDescription
    ]
    allocations: _containers.RepeatedCompositeFieldContainer[Allocation]
    clock: int
    def __init__(
        self,
        function_executors: _Optional[
            _Iterable[_Union[FunctionExecutorDescription, _Mapping]]
        ] = ...,
        allocations: _Optional[_Iterable[_Union[Allocation, _Mapping]]] = ...,
        clock: _Optional[int] = ...,
    ) -> None: ...

class ExecutionPlanUpdate(_message.Message):
    __slots__ = ("function_call", "reduce")
    FUNCTION_CALL_FIELD_NUMBER: _ClassVar[int]
    REDUCE_FIELD_NUMBER: _ClassVar[int]
    function_call: FunctionCall
    reduce: ReduceOp
    def __init__(
        self,
        function_call: _Optional[_Union[FunctionCall, _Mapping]] = ...,
        reduce: _Optional[_Union[ReduceOp, _Mapping]] = ...,
    ) -> None: ...

class FunctionCall(_message.Message):
    __slots__ = ("id", "target", "args", "call_metadata")
    ID_FIELD_NUMBER: _ClassVar[int]
    TARGET_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    CALL_METADATA_FIELD_NUMBER: _ClassVar[int]
    id: str
    target: FunctionRef
    args: _containers.RepeatedCompositeFieldContainer[FunctionArg]
    call_metadata: bytes
    def __init__(
        self,
        id: _Optional[str] = ...,
        target: _Optional[_Union[FunctionRef, _Mapping]] = ...,
        args: _Optional[_Iterable[_Union[FunctionArg, _Mapping]]] = ...,
        call_metadata: _Optional[bytes] = ...,
    ) -> None: ...

class FunctionArg(_message.Message):
    __slots__ = ("function_call_id", "inline_data")
    FUNCTION_CALL_ID_FIELD_NUMBER: _ClassVar[int]
    INLINE_DATA_FIELD_NUMBER: _ClassVar[int]
    function_call_id: str
    inline_data: DataPayload
    def __init__(
        self,
        function_call_id: _Optional[str] = ...,
        inline_data: _Optional[_Union[DataPayload, _Mapping]] = ...,
    ) -> None: ...

class ReduceOp(_message.Message):
    __slots__ = ("id", "collection", "reducer", "call_metadata")
    ID_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_FIELD_NUMBER: _ClassVar[int]
    REDUCER_FIELD_NUMBER: _ClassVar[int]
    CALL_METADATA_FIELD_NUMBER: _ClassVar[int]
    id: str
    collection: _containers.RepeatedCompositeFieldContainer[FunctionArg]
    reducer: FunctionRef
    call_metadata: bytes
    def __init__(
        self,
        id: _Optional[str] = ...,
        collection: _Optional[_Iterable[_Union[FunctionArg, _Mapping]]] = ...,
        reducer: _Optional[_Union[FunctionRef, _Mapping]] = ...,
        call_metadata: _Optional[bytes] = ...,
    ) -> None: ...

class ExecutionPlanUpdates(_message.Message):
    __slots__ = ("updates", "root_function_call_id")
    UPDATES_FIELD_NUMBER: _ClassVar[int]
    ROOT_FUNCTION_CALL_ID_FIELD_NUMBER: _ClassVar[int]
    updates: _containers.RepeatedCompositeFieldContainer[ExecutionPlanUpdate]
    root_function_call_id: str
    def __init__(
        self,
        updates: _Optional[_Iterable[_Union[ExecutionPlanUpdate, _Mapping]]] = ...,
        root_function_call_id: _Optional[str] = ...,
    ) -> None: ...

class AllocationResult(_message.Message):
    __slots__ = (
        "function",
        "allocation_id",
        "function_call_id",
        "request_id",
        "outcome_code",
        "failure_reason",
        "value",
        "updates",
        "request_error",
        "execution_duration_ms",
    )
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    ALLOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_CALL_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    OUTCOME_CODE_FIELD_NUMBER: _ClassVar[int]
    FAILURE_REASON_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    UPDATES_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ERROR_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_DURATION_MS_FIELD_NUMBER: _ClassVar[int]
    function: FunctionRef
    allocation_id: str
    function_call_id: str
    request_id: str
    outcome_code: AllocationOutcomeCode
    failure_reason: AllocationFailureReason
    value: DataPayload
    updates: ExecutionPlanUpdates
    request_error: DataPayload
    execution_duration_ms: int
    def __init__(
        self,
        function: _Optional[_Union[FunctionRef, _Mapping]] = ...,
        allocation_id: _Optional[str] = ...,
        function_call_id: _Optional[str] = ...,
        request_id: _Optional[str] = ...,
        outcome_code: _Optional[_Union[AllocationOutcomeCode, str]] = ...,
        failure_reason: _Optional[_Union[AllocationFailureReason, str]] = ...,
        value: _Optional[_Union[DataPayload, _Mapping]] = ...,
        updates: _Optional[_Union[ExecutionPlanUpdates, _Mapping]] = ...,
        request_error: _Optional[_Union[DataPayload, _Mapping]] = ...,
        execution_duration_ms: _Optional[int] = ...,
    ) -> None: ...

class FunctionCallRequest(_message.Message):
    __slots__ = (
        "parent_request_id",
        "function",
        "inputs",
        "call_metadata",
        "source_function_call_id",
    )
    PARENT_REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    CALL_METADATA_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FUNCTION_CALL_ID_FIELD_NUMBER: _ClassVar[int]
    parent_request_id: str
    function: FunctionRef
    inputs: _containers.RepeatedCompositeFieldContainer[DataPayload]
    call_metadata: bytes
    source_function_call_id: str
    def __init__(
        self,
        parent_request_id: _Optional[str] = ...,
        function: _Optional[_Union[FunctionRef, _Mapping]] = ...,
        inputs: _Optional[_Iterable[_Union[DataPayload, _Mapping]]] = ...,
        call_metadata: _Optional[bytes] = ...,
        source_function_call_id: _Optional[str] = ...,
    ) -> None: ...

class FunctionCallResult(_message.Message):
    __slots__ = ("output", "exception")
    OUTPUT_FIELD_NUMBER: _ClassVar[int]
    EXCEPTION_FIELD_NUMBER: _ClassVar[int]
    output: DataPayload
    exception: str
    def __init__(
        self,
        output: _Optional[_Union[DataPayload, _Mapping]] = ...,
        exception: _Optional[str] = ...,
    ) -> None: ...

class FunctionCallResponse(_message.Message):
    __slots__ = ("update", "result")
    UPDATE_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    update: str
    result: FunctionCallResult
    def __init__(
        self,
        update: _Optional[str] = ...,
        result: _Optional[_Union[FunctionCallResult, _Mapping]] = ...,
    ) -> None: ...



import prometheus_client

from ..monitoring.metrics import latency_metric_for_fast_operation

metric_grpc_server_channel_creations = prometheus_client.Counter(
    "grpc_server_channel_creations",
    "Number of times a channel to gRPC Server was created",
)
metric_grpc_server_channel_creation_retries = prometheus_client.Counter(
    "grpc_server_channel_creation_retries",
    "Number of retries during a channel creation to gRPC Server",
)
metric_grpc_server_channel_creation_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "grpc_server_channel_creation",
        "gRPC server channel creation",
    )
)



from .allocation_output import AllocationOutput
from .function_executor_controller import FunctionExecutorController
from .loggers import allocation_logger, function_executor_logger
from .message_validators import (
    validate_allocation,
    validate_function_executor_description,
)

__all__ = [
    "function_executor_logger",
    "allocation_logger",
    "validate_function_executor_description",
    "validate_allocation",
    "FunctionExecutorController",
    "AllocationOutput",
]



import prometheus_client

from indexify.executor.monitoring.metrics import latency_metric_for_fast_operation

metric_allocation_preparations: prometheus_client.Counter = prometheus_client.Counter(
    "allocation_preparations",
    "Number of allocation preparations for execution",
)
metric_allocation_preparation_errors: prometheus_client.Counter = (
    prometheus_client.Counter(
        "allocation_preparation_errors",
        "Number of allocation preparation errors",
    )
)
metric_allocation_preparation_latency: prometheus_client.Histogram = (
    latency_metric_for_fast_operation(
        "allocation_preparation", "allocation preparation for execution"
    )
)
metric_allocations_getting_prepared: prometheus_client.Gauge = prometheus_client.Gauge(
    "allocations_getting_prepared",
    "Number of allocations currently getting prepared for execution",
)



import prometheus_client

from ...monitoring.metrics import latency_metric_for_fast_operation

# This file contains all metrics used by HealthChecker.

metric_failed_health_checks = prometheus_client.Counter(
    "function_executor_failed_health_checks",
    "Number of health checks that were not successful",
)
metric_health_check_latency = latency_metric_for_fast_operation(
    "function_executor_health_check",
    "Function Executor health check",
)



import asyncio
import time
from typing import Any, List

from tensorlake.function_executor.proto.function_executor_pb2 import (
    BLOB,
    BLOBChunk,
    FunctionInputs,
    SerializedObjectInsideBLOB,
)

from indexify.executor.blob_store.blob_store import BLOBStore
from indexify.proto.executor_api_pb2 import Allocation, DataPayload

from .allocation_info import AllocationInfo
from .allocation_input import AllocationInput
from .downloads import serialized_object_manifest_from_data_payload_proto
from .events import AllocationPreparationFinished
from .metrics.prepare_allocation import (
    metric_allocation_preparation_errors,
    metric_allocation_preparation_latency,
    metric_allocation_preparations,
    metric_allocations_getting_prepared,
)

# The following constants are subject to S3 limits,
# see https://docs.aws.amazon.com/AmazonS3/latest/userguide/qfacts.html.
#
# 7 days - max presigned URL validity duration and limit on max function duration
_MAX_PRESIGNED_URI_EXPIRATION_SEC: int = 7 * 24 * 60 * 60
# This chunk size gives the best performance with S3. Based on our benchmarking.
_BLOB_OPTIMAL_CHUNK_SIZE_BYTES: int = 100 * 1024 * 1024  # 100 MB
# Max output size with optimal chunks is 100 * 100 MB = 10 GB.
# Each chunk requires a separate S3 presign operation, so we limit the number of optimal chunks to 100.
# S3 presign operations are local, it typically takes 30 ms per 100 URLs.
_OUTPUT_BLOB_OPTIMAL_CHUNKS_COUNT: int = 100
# This chunk size gives ~20% slower performance with S3 compared to optimal. Based on our benchmarking.
_OUTPUT_BLOB_SLOWER_CHUNK_SIZE_BYTES: int = 1 * 1024 * 1024 * 1024  # 1 GB
# Max output size with slower chunks is 100 * 1 GB = 100 GB.
_OUTPUT_BLOB_SLOWER_CHUNKS_COUNT: int = 100
# Request error output is using a single chunk.
_REQUEST_ERROR_MAX_SIZE_BYTES: int = 10 * 1024 * 1024  # 10 MB


async def prepare_allocation(
    alloc_info: AllocationInfo, blob_store: BLOBStore, logger: Any
) -> AllocationPreparationFinished:
    """Prepares the allocation for execution.

    If successful then the allocation is runnable.
    Doesn't raise any exceptions.
    """
    logger = logger.bind(module=__name__)
    start_time = time.monotonic()
    try:
        with (
            metric_allocation_preparation_errors.count_exceptions(),
            metric_allocations_getting_prepared.track_inprogress(),
            metric_allocation_preparation_latency.time(),
        ):
            metric_allocation_preparations.inc()
            alloc_info.input = await _prepare_alloc_input(
                alloc_info=alloc_info,
                blob_store=blob_store,
                logger=logger,
            )
            logger.info(
                "allocation was prepared for execution",
                duration=time.monotonic() - start_time,
            )
            return AllocationPreparationFinished(
                alloc_info=alloc_info,
                is_success=True,
            )
    except asyncio.CancelledError:
        return AllocationPreparationFinished(alloc_info=alloc_info, is_success=False)
    except BaseException as e:
        logger.error(
            "failed to prepare allocation for execution",
            exc_info=e,
            duration=time.monotonic() - start_time,
        )
        return AllocationPreparationFinished(alloc_info=alloc_info, is_success=False)


async def _prepare_alloc_input(
    alloc_info: AllocationInfo, blob_store: BLOBStore, logger: Any
) -> AllocationInput:
    """Prepares the alloc for execution.

    Raises an exception on error.
    """
    alloc: Allocation = alloc_info.allocation
    function_outputs_blob_uri: str = (
        f"{alloc.output_payload_uri_prefix}.{alloc_info.allocation.allocation_id}.output"
    )
    request_error_blob_uri: str = (
        f"{alloc.request_error_payload_uri_prefix}.{alloc.request_id}.req_error"
    )

    # The uploads are completed when finalizing the alloc.
    function_outputs_blob_upload_id: str | None = None
    request_error_blob_upload_id: str | None = None

    try:
        function_outputs_blob_upload_id = await blob_store.create_multipart_upload(
            uri=function_outputs_blob_uri,
            logger=logger,
        )
        request_error_blob_upload_id = await blob_store.create_multipart_upload(
            uri=request_error_blob_uri,
            logger=logger,
        )
    except BaseException:
        if function_outputs_blob_upload_id is not None:
            await blob_store.abort_multipart_upload(
                uri=function_outputs_blob_uri,
                upload_id=function_outputs_blob_upload_id,
                logger=logger,
            )
        if request_error_blob_upload_id is not None:
            await blob_store.abort_multipart_upload(
                uri=request_error_blob_uri,
                upload_id=request_error_blob_upload_id,
                logger=logger,
            )
        raise

    args: List[SerializedObjectInsideBLOB] = []
    arg_blobs: List[BLOB] = []
    for arg in alloc.args:
        arg: DataPayload
        args.append(_to_serialized_object_inside_blob(arg))
        arg_blob = await _presign_function_arg_blob(
            arg=arg,
            blob_store=blob_store,
            logger=logger,
        )
        arg_blobs.append(arg_blob)

    return AllocationInput(
        function_inputs=FunctionInputs(
            args=args,
            arg_blobs=arg_blobs,
            function_outputs_blob=await _presign_function_outputs_blob(
                uri=function_outputs_blob_uri,
                upload_id=function_outputs_blob_upload_id,
                blob_store=blob_store,
                logger=logger,
            ),
            request_error_blob=await _presign_request_error_blob(
                uri=request_error_blob_uri,
                upload_id=request_error_blob_upload_id,
                blob_store=blob_store,
                logger=logger,
            ),
            function_call_metadata=alloc.function_call_metadata,
        ),
        function_outputs_blob_uri=function_outputs_blob_uri,
        function_outputs_blob_upload_id=function_outputs_blob_upload_id,
        request_error_blob_uri=request_error_blob_uri,
        request_error_blob_upload_id=request_error_blob_upload_id,
    )


async def _presign_function_arg_blob(
    arg: DataPayload, blob_store: BLOBStore, logger: Any
) -> BLOB:
    get_blob_uri: str = await blob_store.presign_get_uri(
        uri=arg.uri,
        expires_in_sec=_MAX_PRESIGNED_URI_EXPIRATION_SEC,
        logger=logger,
    )
    chunks: List[BLOBChunk] = []

    while len(chunks) * _BLOB_OPTIMAL_CHUNK_SIZE_BYTES < arg.size:
        chunks.append(
            BLOBChunk(
                uri=get_blob_uri,  # The URI allows to read any byte range in the BLOB.
                size=_BLOB_OPTIMAL_CHUNK_SIZE_BYTES,
                # ETag is only set by FE when returning BLOBs to us
            )
        )

    return BLOB(
        chunks=chunks,
    )


async def _presign_function_outputs_blob(
    uri: str, upload_id: str, blob_store: BLOBStore, logger: Any
) -> BLOB:
    """Presigns the output blob for the allocation."""
    chunks: List[BLOBChunk] = []

    while len(chunks) != (
        _OUTPUT_BLOB_OPTIMAL_CHUNKS_COUNT + _OUTPUT_BLOB_SLOWER_CHUNKS_COUNT
    ):
        upload_chunk_uri: str = await blob_store.presign_upload_part_uri(
            uri=uri,
            part_number=len(chunks) + 1,
            upload_id=upload_id,
            expires_in_sec=_MAX_PRESIGNED_URI_EXPIRATION_SEC,
            logger=logger,
        )

        chunk_size: int = (
            _BLOB_OPTIMAL_CHUNK_SIZE_BYTES
            if len(chunks) < _OUTPUT_BLOB_OPTIMAL_CHUNKS_COUNT
            else _OUTPUT_BLOB_SLOWER_CHUNK_SIZE_BYTES
        )
        chunks.append(
            BLOBChunk(
                uri=upload_chunk_uri,
                size=chunk_size,
                # ETag is only set by FE when returning BLOBs to us
            )
        )

    return BLOB(
        chunks=chunks,
    )


async def _presign_request_error_blob(
    uri: str, upload_id: str, blob_store: BLOBStore, logger: Any
) -> BLOB:
    """Presigns the output blob for the request error."""
    upload_chunk_uri: str = await blob_store.presign_upload_part_uri(
        uri=uri,
        part_number=1,
        upload_id=upload_id,
        expires_in_sec=_MAX_PRESIGNED_URI_EXPIRATION_SEC,
        logger=logger,
    )
    return BLOB(
        chunks=[
            BLOBChunk(
                uri=upload_chunk_uri,
                size=_REQUEST_ERROR_MAX_SIZE_BYTES,
                # ETag is only set by FE when returning BLOBs to us
            )
        ]
    )


def _to_serialized_object_inside_blob(
    data_payload: DataPayload,
) -> SerializedObjectInsideBLOB:
    return SerializedObjectInsideBLOB(
        manifest=serialized_object_manifest_from_data_payload_proto(data_payload),
        offset=data_payload.offset,
    )



import subprocess
from dataclasses import dataclass
from enum import Enum
from typing import Any, List


# Only NVIDIA GPUs currently supported in Tensorlake SDK are listed here.
# GPU models coming with multiple memory sizes have a different enum value per memory size.
class NVIDIA_GPU_MODEL(str, Enum):
    UNKNOWN = "UNKNOWN"
    A100_40GB = "A100-40GB"
    A100_80GB = "A100-80GB"
    H100_80GB = "H100-80GB"
    TESLA_T4 = "T4"
    A6000 = "A6000"
    A10 = "A10"


@dataclass
class NvidiaGPUInfo:
    index: str
    uuid: str
    product_name: str  # The official product name.
    model: NVIDIA_GPU_MODEL


def nvidia_gpus_are_available() -> bool:
    try:
        result: subprocess.CompletedProcess = subprocess.run(
            ["nvidia-smi"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        return result.returncode == 0
    except Exception:
        return False


def fetch_nvidia_gpu_infos(logger: Any) -> List[NvidiaGPUInfo]:
    logger = logger.bind(module=__name__)
    logger.info("Fetching GPU information")

    try:
        result: subprocess.CompletedProcess = subprocess.run(
            ["nvidia-smi", "--query-gpu=index,name,uuid", "--format=csv,noheader"],
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        logger.error("Failed to fetch GPU information", exc_info=e)
        raise

    infos: List[NvidiaGPUInfo] = []
    for line in result.stdout.splitlines():
        # Example:
        # nvidia-smi --query-gpu=index,name,uuid --format=csv,noheader
        # 0, NVIDIA A100-SXM4-80GB, GPU-89fdc1e1-18b2-f499-c12b-82bcb9bfb3fa
        # 1, NVIDIA A100-PCIE-40GB, GPU-e9c9aa65-bff3-405a-ab7c-dc879cc88169
        # 2, NVIDIA H100 80GB HBM3, GPU-8c35f4c9-4dff-c9a2-866f-afb5d82e1dd7
        # 3, Tesla T4, GPU-2a7fadae-a692-1c44-2c57-6645a0d117e4
        # 4, NVIDIA RTX A6000, GPU-efe4927a-743f-e4cc-28bb-da604f545b6d
        # 5, NVIDIA A10, GPU-12463b8c-40bb-7322-6c7a-ef48bd7bd39b
        parts = line.split(",")
        index = parts[0].strip()
        product_name = parts[1].strip()
        uuid = parts[2].strip()

        model = _product_name_to_model(product_name)
        if model == NVIDIA_GPU_MODEL.UNKNOWN:
            logger.warning(
                "Unknown GPU model was detected, ignoring", nvidia_smi_output=line
            )
        infos.append(
            NvidiaGPUInfo(
                index=index, uuid=uuid, product_name=product_name, model=model
            )
        )

    return infos


def _product_name_to_model(product_name: str) -> NVIDIA_GPU_MODEL:
    if product_name.startswith("NVIDIA A100") and product_name.endswith("80GB"):
        return NVIDIA_GPU_MODEL.A100_80GB
    if product_name.startswith("NVIDIA A100") and product_name.endswith("40GB"):
        return NVIDIA_GPU_MODEL.A100_40GB
    elif product_name.startswith("NVIDIA H100") and "80GB" in product_name:
        return NVIDIA_GPU_MODEL.H100_80GB
    elif product_name.startswith("Tesla T4"):
        return NVIDIA_GPU_MODEL.TESLA_T4
    elif product_name.startswith("NVIDIA RTX A6000"):
        return NVIDIA_GPU_MODEL.A6000
    elif product_name.startswith("NVIDIA A10"):
        return NVIDIA_GPU_MODEL.A10
    else:
        return NVIDIA_GPU_MODEL.UNKNOWN



import asyncio
import math
import time
from collections.abc import Coroutine
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List

from tensorlake.function_executor.proto.function_executor_pb2 import (
    ExecutionPlanUpdate as FEExecutionPlanUpdate,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    ExecutionPlanUpdates as FEExecutionPlanUpdates,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    FunctionArg as FEFunctionArg,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    FunctionCall as FEFunctionCall,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    FunctionRef as FEFunctionRef,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    ReduceOp as FEReduceOp,
)
from tensorlake.function_executor.proto.function_executor_pb2 import (
    SerializedObjectEncoding,
    SerializedObjectInsideBLOB,
)

from indexify.executor.blob_store.blob_store import BLOBStore
from indexify.executor.function_executor.function_executor import FunctionExecutor
from indexify.executor.function_executor.health_checker import HealthCheckResult
from indexify.executor.function_executor.server.function_executor_server_factory import (
    FunctionExecutorServerFactory,
)
from indexify.executor.state_reporter import ExecutorStateReporter
from indexify.proto.executor_api_pb2 import (
    Allocation,
    AllocationResult,
    DataPayload,
    DataPayloadEncoding,
    ExecutionPlanUpdate,
    ExecutionPlanUpdates,
    FunctionArg,
    FunctionCall,
    FunctionExecutorDescription,
    FunctionExecutorState,
    FunctionExecutorStatus,
    FunctionExecutorTerminationReason,
    FunctionRef,
    ReduceOp,
)

from .allocation_info import AllocationInfo
from .allocation_input import AllocationInput
from .allocation_output import AllocationOutput
from .completed_allocation_metrics import emit_completed_allocation_metrics
from .create_function_executor import create_function_executor
from .debug_event_loop import (
    debug_print_adding_event,
    debug_print_events,
    debug_print_processing_event,
)
from .events import (
    AllocationExecutionFinished,
    AllocationFinalizationFinished,
    AllocationPreparationFinished,
    BaseEvent,
    EventType,
    FunctionExecutorCreated,
    FunctionExecutorTerminated,
    ScheduleAllocationExecution,
    ShutdownInitiated,
)
from .finalize_allocation import finalize_allocation
from .loggers import allocation_logger, function_executor_logger
from .metrics.function_executor_controller import (
    METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_NOT_STARTED,
    METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_RUNNING,
    METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_STARTING_UP,
    METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATED,
    METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATING,
    METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_UNKNOWN,
    metric_allocations_fetched,
    metric_control_loop_handle_event_latency,
    metric_function_executors_with_state,
    metric_runnable_allocations,
    metric_runnable_allocations_per_function_name,
    metric_schedule_allocation_latency,
)
from .prepare_allocation import prepare_allocation
from .run_allocation import run_allocation_on_function_executor
from .terminate_function_executor import terminate_function_executor


# Actual FE controller states, they are a bit different from statuses reported to the Server.
# All the valid state transitions are forward only (can skip multiple states in a row).
class _FE_CONTROLLER_STATE(Enum):
    NOT_STARTED = 1
    STARTING_UP = 2
    RUNNING = 3
    TERMINATING = 4
    TERMINATED = 5


class FunctionExecutorController:
    def __init__(
        self,
        executor_id: str,
        function_executor_description: FunctionExecutorDescription,
        function_executor_server_factory: FunctionExecutorServerFactory,
        state_reporter: ExecutorStateReporter,
        blob_store: BLOBStore,
        base_url: str,
        config_path: str,
        cache_path: Path,
        logger: Any,
    ):
        """Initializes the FunctionExecutorController.

        The supplied FunctionExecutorDescription must be already validated by the caller
        using validate_function_executor_description().
        """
        self._executor_id: str = executor_id
        self._fe_description: FunctionExecutorDescription = (
            function_executor_description
        )
        self._fe_server_factory: FunctionExecutorServerFactory = (
            function_executor_server_factory
        )
        self._state_reporter: ExecutorStateReporter = state_reporter
        self._blob_store: BLOBStore = blob_store
        self._base_url: str = base_url
        self._config_path: str = config_path
        self._cache_path: Path = cache_path
        self._logger: Any = function_executor_logger(
            function_executor_description, logger.bind(module=__name__)
        )
        self._destroy_lock: asyncio.Lock = asyncio.Lock()
        # Mutable state. No lock needed as it's modified by async tasks running in the same event loop.
        self._fe: FunctionExecutor | None = None
        self._internal_state = _FE_CONTROLLER_STATE.NOT_STARTED
        metric_function_executors_with_state.labels(
            state=_to_fe_state_metric_label(self._internal_state, self._logger)
        ).inc()
        self._reported_state: FunctionExecutorState = FunctionExecutorState(
            description=function_executor_description,
            status=FunctionExecutorStatus.FUNCTION_EXECUTOR_STATUS_UNKNOWN,
        )
        # Ordered list of events to be processed by the control loop.
        self._events: List[BaseEvent] = []
        # Asyncio event used to notify the control loop that there are new events to process.
        self._event_added: asyncio.Event = asyncio.Event()
        # Control loop asyncio task.
        self._control_loop_aio_task: asyncio.Task | None = None
        # aio tasks spawned by the control loop.
        self._running_aio_tasks: List[asyncio.Task] = []
        # All allocations assigned to FE, Allocation ID -> AllocationInfo.
        self._allocations: Dict[str, AllocationInfo] = {}
        # Allocations prepared for execution on FE.
        self._runnable_allocations: List[AllocationInfo] = []
        # Allocations currently running on the FE.
        self._running_allocations: List[AllocationInfo] = []

    def function_executor_id(self) -> str:
        return self._fe_description.id

    def add_allocation(self, allocation: Allocation) -> None:
        """Adds an allocation to the Function Executor.

        Not blocking. Never raises exceptions.
        """
        logger = allocation_logger(allocation, self._logger)
        if self.has_allocation(allocation.allocation_id):
            logger.warning(
                "attempted to add already added allocation to Function Executor",
            )
            return

        metric_allocations_fetched.inc()
        alloc_info: AllocationInfo = AllocationInfo(
            allocation=allocation,
            allocation_timeout_ms=self._fe_description.allocation_timeout_ms,
            start_time=time.monotonic(),
        )
        self._allocations[allocation.allocation_id] = alloc_info
        next_aio = prepare_allocation(
            alloc_info=alloc_info,
            blob_store=self._blob_store,
            logger=logger,
        )
        self._spawn_aio_for_allocation(
            aio=next_aio,
            alloc_info=alloc_info,
            on_exception=AllocationPreparationFinished(
                alloc_info=alloc_info, is_success=False
            ),
        )

    def has_allocation(self, allocation_id: str) -> bool:
        """Checks if the Function Executor has an allocation with the given ID.

        Not blocking. Never raises exceptions.
        """
        return allocation_id in self._allocations

    def allocation_ids(self) -> List[str]:
        """Returns the list of allocation IDs known to the Function Executor.

        Not blocking. Never raises exceptions.
        """
        return list(self._allocations.keys())

    def remove_allocation(self, allocation_id: str) -> None:
        """Removes the allocation from the Function Executor.

        Cancels the allocation if it's in progress. Just removes the allocation if it was already
        completed. The cancellation is asynchronous and might take a while to complete.
        Until the cancellation is complete, the allocation won't be removed from the Function Executor.
        Not blocking. Never raises exceptions.
        """
        if not self.has_allocation(allocation_id):
            self._logger.warning(
                "attempted to cancel an allocation that is not known to the Function Executor",
                allocation_id=allocation_id,
            )
            return

        alloc_info: AllocationInfo = self._allocations.pop(allocation_id)
        if alloc_info.is_completed:
            return  # Server processed the completed alloc outputs, we can forget it now.

        # Alloc cancellation is required as the alloc is not completed yet.
        logger = allocation_logger(alloc_info.allocation, self._logger)
        alloc_info.is_cancelled = True
        logger.info(
            "cancelling allocation",
        )
        if alloc_info.aio_task is not None:
            alloc_info.aio_task.cancel()

    def startup(self) -> None:
        """Starts up the Function Executor and prepares it to run allocations.

        Not blocking. Never raises exceptions."""
        if self._internal_state != _FE_CONTROLLER_STATE.NOT_STARTED:
            self._logger.warning(
                "function executor state is not NOT_STARTED, ignoring startup call",
                internal_state=self._internal_state.name,
            )
            return

        self._control_loop_aio_task = asyncio.create_task(
            self._control_loop(),
            name="function executor control loop",
        )
        self._update_internal_state(_FE_CONTROLLER_STATE.STARTING_UP)
        self._update_reported_state(
            FunctionExecutorState(
                description=self._fe_description,
                status=FunctionExecutorStatus.FUNCTION_EXECUTOR_STATUS_PENDING,
            )
        )
        next_aio = create_function_executor(
            function_executor_description=self._fe_description,
            function_executor_server_factory=self._fe_server_factory,
            blob_store=self._blob_store,
            executor_id=self._executor_id,
            base_url=self._base_url,
            config_path=self._config_path,
            cache_path=self._cache_path,
            logger=self._logger,
        )
        self._spawn_aio_for_fe(
            aio=next_aio,
            on_exception=FunctionExecutorCreated(
                function_executor=None,
                fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_INTERNAL_ERROR,
            ),
        )

    async def shutdown(self) -> None:
        """Shutsdown the Function Executor and frees all of its resources.

        No alloc outcomes and outputs are getting reported to Server after this call.
        Doesn't raise any exceptions. Blocks until the shutdown is complete. Idempotent.
        """
        self._add_event(ShutdownInitiated(), source="shutdown")
        try:
            await self._control_loop_aio_task
        except asyncio.CancelledError:
            pass  # Expected exception on shutdown
        except Exception as e:
            self._logger.error(
                "function executor controller control loop task raised unexpected exception",
                exc_info=e,
            )

    def _update_internal_state(self, new_state: _FE_CONTROLLER_STATE) -> None:
        """Updates the internal state of the Function Executor Controller.

        Not blocking. Never raises exceptions."""
        old_state: _FE_CONTROLLER_STATE = self._internal_state
        self._internal_state = new_state

        self._logger.info(
            "function executor internal state changed",
            old_state=old_state.name,
            new_state=new_state.name,
        )

        metric_function_executors_with_state.labels(
            state=_to_fe_state_metric_label(old_state, self._logger)
        ).dec()
        metric_function_executors_with_state.labels(
            state=_to_fe_state_metric_label(new_state, self._logger)
        ).inc()

    def _update_reported_state(
        self,
        new_state: FunctionExecutorState,
    ) -> None:
        """Sets new Function Executor state and reports it to the Server.

        Not blocking. Never raises exceptions."""
        old_state: FunctionExecutorState = self._reported_state
        self._reported_state = new_state

        self._logger.info(
            "function executor grpc status changed",
            old_status=FunctionExecutorStatus.Name(old_state.status),
            new_status=FunctionExecutorStatus.Name(new_state.status),
            termination_reason=_termination_reason_to_short_name(
                new_state.termination_reason
            ),
        )

        self._state_reporter.update_function_executor_state(new_state)
        # Report the status change to the Server asap to reduce latency in the system.
        self._state_reporter.schedule_state_report()

    async def _control_loop(self) -> None:
        """Runs control loop that coordinates all the work done by the Function Executor.

        Doesn't raise any Exceptions.
        """
        self._logger.info("function executor controller control loop started")

        while True:
            await self._event_added.wait()
            self._event_added.clear()

            while self._events:
                event: BaseEvent = self._events.pop(0)
                debug_print_processing_event(event, self._logger)

                try:
                    if event.event_type == EventType.SHUTDOWN_INITIATED:
                        return await self._shutdown(event)

                    with metric_control_loop_handle_event_latency.time():
                        self._handle_event(event)
                except BaseException as e:
                    # None of the event handlers should raise exceptions, but still catch all exceptions to ensure
                    # that the control loop doesn't crash if an unexpected exception happen.
                    self._logger.error(
                        "unexpected exception in function executor controller control loop",
                        exc_info=e,
                        event_type=event.event_type.name,
                    )
                    if event.event_type == EventType.SHUTDOWN_INITIATED:
                        return  # Unexpected exception during shutdown, should return anyway.

    def _handle_event(self, event: BaseEvent) -> None:
        """Handles the event.

        Doesn't raise any exceptions. Doesn't block.
        """
        if event.event_type == EventType.FUNCTION_EXECUTOR_CREATED:
            return self._handle_event_function_executor_created(event)
        elif event.event_type == EventType.FUNCTION_EXECUTOR_TERMINATED:
            return self._handle_event_function_executor_terminated(event)
        elif event.event_type == EventType.ALLOCATION_PREPARATION_FINISHED:
            return self._handle_event_allocation_preparation_finished(event)
        elif event.event_type == EventType.SCHEDULE_ALLOCATION_EXECUTION:
            return self._handle_event_schedule_allocation_execution(event)
        elif event.event_type == EventType.ALLOCATION_EXECUTION_FINISHED:
            return self._handle_event_allocation_execution_finished(event)
        elif event.event_type == EventType.ALLOCATION_FINALIZATION_FINISHED:
            return self._handle_event_allocation_finalization_finished(event)

        self._logger.warning(
            "unexpected event type received", event_type=event.event_type.name
        )

    def _add_event(self, event: BaseEvent, source: str) -> None:
        """Adds an event to the list of events to be processed by the control loop.

        Doesn't raise any exceptions. Doesn't block."""
        debug_print_adding_event(event=event, source=source, logger=self._logger)
        self._events.append(event)
        self._event_added.set()

    def _spawn_aio_for_allocation(
        self,
        aio: Coroutine[Any, Any, BaseEvent],
        alloc_info: AllocationInfo,
        on_exception: BaseEvent,
    ) -> None:
        self._spawn_aio(
            aio=aio,
            alloc_info=alloc_info,
            on_exception=on_exception,
            logger=allocation_logger(alloc_info.allocation, self._logger),
        )

    def _spawn_aio_for_fe(
        self, aio: Coroutine[Any, Any, BaseEvent], on_exception: BaseEvent
    ) -> None:
        self._spawn_aio(
            aio=aio,
            alloc_info=None,
            on_exception=on_exception,
            logger=self._logger,
        )

    def _spawn_aio(
        self,
        aio: Coroutine[Any, Any, BaseEvent],
        alloc_info: AllocationInfo | None,
        on_exception: BaseEvent,
        logger: Any,
    ) -> None:
        """Spawns an aio task for the supplied coroutine.

        The coroutine should return an event that will be added to the FE controller events.
        The coroutine should not raise any exceptions including BaseException.
        on_exception event will be added to the FE controller events if the aio task raises an unexpected exception.
        on_exception is required to not silently stall the task processing due to an unexpected exception.
        If alloc_info is not None, the aio task will be associated with the alloc_info while the aio task is running.
        Doesn't raise any exceptions. Doesn't block.
        Use `_spawn_aio_for_task_alloc` and `_spawn_aio_for_fe` instead of directly calling this method.
        """

        aio_task_name: str = str(aio)
        # Wrap the coroutine into aio task to disable warning "coroutine was never awaited" when the task is cancelled.
        aio: asyncio.Task = asyncio.create_task(aio, name=aio_task_name)

        async def coroutine_wrapper() -> None:
            try:
                self._add_event(await aio, source=aio_task_name)
            except asyncio.CancelledError:
                # Workaround for scenario when coroutine_wrapper gets cancelled at `await aio` before aio starts.
                # In this case aio doesn't handle the cancellation and doesn't return the right event.
                # A fix for this is to cancel aio instead of coroutine_wrapper. We'll need to keep
                # references to both coroutine_wrapper and aio, cause event loop uses weak references to
                # tasks. Not doing this for now. Using on_exception is good enough because not started aios don't
                # need to do anything special on cancellation.
                self._add_event(on_exception, source=aio_task_name)
            except BaseException as e:
                logger.error(
                    "unexpected exception in aio task",
                    exc_info=e,
                    aio_task_name=aio_task_name,
                )
                self._add_event(on_exception, source=aio_task_name)
            finally:
                if alloc_info is not None:
                    alloc_info.aio_task = None
                self._running_aio_tasks.remove(asyncio.current_task())

        aio_wrapper_task: asyncio.Task = asyncio.create_task(
            coroutine_wrapper(),
            name=f"function executor controller aio task '{aio_task_name}'",
        )
        self._running_aio_tasks.append(aio_wrapper_task)
        if alloc_info is not None:
            alloc_info.aio_task = aio_wrapper_task

    # Event handlers for the events added to the control loop.
    # All the event handlers are synchronous and never block on any long running operations.

    def _handle_event_function_executor_created(
        self, event: FunctionExecutorCreated
    ) -> None:
        """Handles the startup finished event.

        Doesn't raise any exceptions. Doesn't block.
        """
        if event.function_executor is None:
            # Server needs to increment attempts counter for all the allocs that were pending while FE was starting up.
            # This prevents infinite retries if FEs consistently fail to start up.
            # The allocations we marked here also need to not used FE terminated failure reason in their outputs
            # because FE terminated means that the allocation wasn't the cause of the FE termination.
            allocation_ids_caused_termination: List[str] = []
            for alloc_info in self._allocations.values():
                alloc_logger = allocation_logger(alloc_info.allocation, self._logger)
                alloc_logger.info(
                    "marking allocation failed on function executor startup failure"
                )
                allocation_ids_caused_termination.append(
                    alloc_info.allocation.allocation_id
                )
                alloc_info.output = AllocationOutput.function_executor_startup_failed(
                    allocation=alloc_info.allocation,
                    fe_termination_reason=event.fe_termination_reason,
                    logger=alloc_logger,
                )
            self._start_termination(
                fe_termination_reason=event.fe_termination_reason,
                allocation_ids_caused_termination=allocation_ids_caused_termination,
            )
            return

        self._fe = event.function_executor
        self._update_internal_state(_FE_CONTROLLER_STATE.RUNNING)
        self._update_reported_state(
            FunctionExecutorState(
                description=self._fe_description,
                status=FunctionExecutorStatus.FUNCTION_EXECUTOR_STATUS_RUNNING,
            )
        )
        # Health checker starts after FE creation and gets automatically stopped on FE destroy.
        self._fe.health_checker().start(self._health_check_failed_callback)
        self._add_event(
            ScheduleAllocationExecution(),
            source="_handle_event_function_executor_created",
        )

    def _handle_event_function_executor_terminated(
        self, event: FunctionExecutorTerminated
    ) -> None:
        """Handles the Function Executor terminated event.

        Doesn't raise any exceptions. Doesn't block.
        """
        if not event.is_success:
            self._logger.error(
                "Function Executor termination failed unexpectedly, this should never happen",
            )

        self._fe = None
        # Set reported status only after the FE got destroyed because Server assumes that all FE resources are freed when the status changes.
        self._update_reported_state(
            FunctionExecutorState(
                description=self._fe_description,
                status=FunctionExecutorStatus.FUNCTION_EXECUTOR_STATUS_TERMINATED,
                termination_reason=event.fe_termination_reason,
                allocation_ids_caused_termination=event.allocation_ids_caused_termination,
            )
        )
        self._update_internal_state(_FE_CONTROLLER_STATE.TERMINATED)

        # Invoke the scheduler so it can fail runnable allocs with FE Terminated error.
        self._add_event(
            ScheduleAllocationExecution(),
            source="_handle_event_function_executor_destroyed",
        )

    async def _health_check_failed_callback(self, result: HealthCheckResult):
        self._logger.error(
            "Function Executor health check failed, terminating Function Executor",
            reason=result.reason,
        )

        self._start_termination(
            fe_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY,
            allocation_ids_caused_termination=[
                alloc_info.allocation.allocation_id
                for alloc_info in self._running_allocations
            ],
        )

    def _handle_event_allocation_preparation_finished(
        self, event: AllocationPreparationFinished
    ) -> None:
        """Handles the allocation preparation finished event.

        Doesn't raise any exceptions. Doesn't block.
        """
        alloc_info: AllocationInfo = event.alloc_info

        if alloc_info.is_cancelled:
            alloc_info.output = AllocationOutput.allocation_cancelled(
                allocation=alloc_info.allocation,
                # Task alloc was never executed
                execution_start_time=None,
                execution_end_time=None,
            )
            self._start_allocation_finalization(alloc_info)
            return

        if not event.is_success:
            # Failed to prepare the alloc inputs.
            alloc_info.output = AllocationOutput.internal_error(
                allocation=alloc_info.allocation,
                # Alloc was never executed
                execution_start_time=None,
                execution_end_time=None,
            )
            self._start_allocation_finalization(alloc_info)
            return

        alloc_info.prepared_time = time.monotonic()
        metric_runnable_allocations.inc()
        metric_runnable_allocations_per_function_name.labels(
            alloc_info.allocation.function.function_name
        ).inc()
        self._runnable_allocations.append(alloc_info)
        self._add_event(
            ScheduleAllocationExecution(),
            source="_handle_event_allocation_preparation_finished",
        )

    def _handle_event_schedule_allocation_execution(
        self, event: ScheduleAllocationExecution
    ) -> None:
        if len(self._runnable_allocations) == 0:
            return

        if self._internal_state not in [
            _FE_CONTROLLER_STATE.RUNNING,
            _FE_CONTROLLER_STATE.TERMINATING,
            _FE_CONTROLLER_STATE.TERMINATED,
        ]:
            return  # Can't progress runnable allocs in the current state.

        if (
            self._internal_state == _FE_CONTROLLER_STATE.RUNNING
            and len(self._running_allocations) == self._fe_description.max_concurrency
        ):
            return

        # Take the next alloc from head to get FIFO order and improve fairness.
        alloc_info: AllocationInfo = self._pop_runnable_allocation()
        # Re-invoke the scheduler later to process the next runnable task if this one can't run on FE.
        self._add_event(
            ScheduleAllocationExecution(),
            source="_handle_event_schedule_allocation_execution",
        )

        if alloc_info.is_cancelled:
            alloc_info.output = AllocationOutput.allocation_cancelled(
                allocation=alloc_info.allocation,
                # Alloc was never executed
                execution_start_time=None,
                execution_end_time=None,
            )
            self._start_allocation_finalization(alloc_info)
        elif self._internal_state in [
            _FE_CONTROLLER_STATE.TERMINATING,
            _FE_CONTROLLER_STATE.TERMINATED,
        ]:
            # The output could be set already by FE startup failure handler.
            if alloc_info.output is None:
                alloc_info.output = AllocationOutput.function_executor_terminated(
                    alloc_info.allocation
                )
            self._start_allocation_finalization(alloc_info)
        elif self._internal_state == _FE_CONTROLLER_STATE.RUNNING:
            self._running_allocations.append(alloc_info)
            next_aio = run_allocation_on_function_executor(
                alloc_info=alloc_info,
                function_executor=self._fe,
                logger=allocation_logger(alloc_info.allocation, self._logger),
            )
            self._spawn_aio_for_allocation(
                aio=next_aio,
                alloc_info=alloc_info,
                on_exception=AllocationExecutionFinished(
                    alloc_info=alloc_info,
                    function_executor_termination_reason=FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_INTERNAL_ERROR,
                ),
            )
        else:
            allocation_logger(alloc_info.allocation, self._logger).error(
                "failed to schedule allocation execution, this should never happen"
            )

    def _pop_runnable_allocation(self) -> AllocationInfo:
        alloc_info: AllocationInfo = self._runnable_allocations.pop(0)
        metric_schedule_allocation_latency.observe(
            time.monotonic() - alloc_info.prepared_time
        )
        metric_runnable_allocations.dec()
        metric_runnable_allocations_per_function_name.labels(
            alloc_info.allocation.function.function_name
        ).dec()
        return alloc_info

    def _handle_event_allocation_execution_finished(
        self, event: AllocationExecutionFinished
    ) -> None:
        """Handles the allocation execution finished event.

        Doesn't raise any exceptions. Doesn't block.
        """
        alloc_info: AllocationInfo = event.alloc_info
        self._running_allocations.remove(alloc_info)

        if event.function_executor_termination_reason is None:
            self._add_event(
                ScheduleAllocationExecution(),
                source="_handle_event_allocation_execution_finished",
            )
        else:
            self._start_termination(
                fe_termination_reason=event.function_executor_termination_reason,
                allocation_ids_caused_termination=[alloc_info.allocation.allocation_id],
            )

        if alloc_info.output is None:
            # `run_allocation_on_function_executor` guarantees that the output is set in
            # all cases including allocation cancellations. If this didn't happen then some
            # internal error occurred in our code.
            alloc_info.output = AllocationOutput.internal_error(
                allocation=alloc_info.allocation,
                execution_start_time=None,
                execution_end_time=None,
            )

        self._start_allocation_finalization(alloc_info)

    def _start_allocation_finalization(self, alloc_info: AllocationInfo) -> None:
        """Starts finalization for the given allocation.

        Doesn't raise any exceptions. Doesn't block.
        alloc_info.output should not be None.
        """
        next_aio = finalize_allocation(
            alloc_info=alloc_info,
            blob_store=self._blob_store,
            logger=allocation_logger(alloc_info.allocation, self._logger),
        )
        self._spawn_aio_for_allocation(
            aio=next_aio,
            alloc_info=alloc_info,
            on_exception=AllocationFinalizationFinished(
                alloc_info=alloc_info, is_success=False
            ),
        )

    def _handle_event_allocation_finalization_finished(
        self, event: AllocationFinalizationFinished
    ) -> None:
        """Handles the allocation finalization finished event.

        Doesn't raise any exceptions. Doesn't block.
        """
        alloc_info: AllocationInfo = event.alloc_info
        if not event.is_success:
            original_alloc_output: AllocationOutput = (
                alloc_info.output
            )  # Never None here
            alloc_info.output = AllocationOutput.internal_error(
                allocation=alloc_info.allocation,
                execution_start_time=original_alloc_output.execution_start_time,
                execution_end_time=original_alloc_output.execution_end_time,
            )

        logger: Any = allocation_logger(alloc_info.allocation, self._logger)
        # Ignore alloc cancellation as it's technically finished at this point.
        alloc_info.is_completed = True
        emit_completed_allocation_metrics(
            alloc_info=alloc_info,
            logger=logger,
        )
        # Reconciler will call .remove_allocation() once Server signals that it processed this update.
        self._state_reporter.add_completed_allocation_result(
            _to_alloc_result_proto(alloc_info, logger)
        )
        self._state_reporter.schedule_state_report()

    def _start_termination(
        self,
        fe_termination_reason: FunctionExecutorTerminationReason,
        allocation_ids_caused_termination: List[str],
    ) -> None:
        """Starts termination of the Function Executor if it's not started yet.

        Doesn't raise any exceptions. Doesn't block.
        """
        if self._internal_state in [
            _FE_CONTROLLER_STATE.TERMINATING,
            _FE_CONTROLLER_STATE.TERMINATED,
        ]:
            # _start_termination() can be called multiple times, e.g. by each failed alloc
            # when the FE is unhealthy. Dedup the calls to keep state machine consistent.
            return

        self._update_internal_state(_FE_CONTROLLER_STATE.TERMINATING)
        next_aio = terminate_function_executor(
            function_executor=self._fe,
            lock=self._destroy_lock,
            fe_termination_reason=fe_termination_reason,
            allocation_ids_caused_termination=allocation_ids_caused_termination,
            logger=self._logger,
        )
        self._spawn_aio_for_fe(
            aio=next_aio,
            on_exception=FunctionExecutorTerminated(
                is_success=False,
                fe_termination_reason=fe_termination_reason,
                allocation_ids_caused_termination=allocation_ids_caused_termination,
            ),
        )

    async def _shutdown(self, event: ShutdownInitiated) -> None:
        """Shuts down the Function Executor and frees all its resources.

        The control loop must be blocked while this method is running.
        The control loop must exit immediately after this method returns.
        Doesn't raise any exceptions.

        Server needs to wait until all the allocations its interested in got their outcomes reported
        before calling the FE shutdown as we don't report anything on FE shutdown.
        """
        self._logger.info("function executor controller shutdown initiated")
        # Control loop is blocked executing this method, no new aio tasks will be spawned concurrently.
        # Create a copy of the running aio tasks because they remove themselves from the list when they finish.
        cancelled_tasks: List[asyncio.Task] = self._running_aio_tasks.copy()
        for cancelled_task in cancelled_tasks:
            cancelled_task.cancel()

        # Await all aio tasks to make sure that nothing is mutating this FE controller state concurrently.
        for cancelled_task in cancelled_tasks:
            try:
                await cancelled_task
            except BaseException:
                # Ignore any errors as we expect them when cancelling tasks.
                # BaseException includes asyncio.CancelledError which is always raised here.
                pass

        # Makes sure we don't run fe destroy concurrently with an event loop task.
        # FE destroy uses asyncio.to_thread() calls so it doesn't get cancelled with all the tasks above.
        async with self._destroy_lock:
            if self._fe is not None:
                self._logger.info(
                    "destroying function executor",
                )
                await self._fe.destroy()

        # Cleanup the metric from this FE.
        metric_function_executors_with_state.labels(
            state=_to_fe_state_metric_label(self._internal_state, self._logger)
        ).dec()

        self._state_reporter.remove_function_executor_state(self.function_executor_id())
        self._state_reporter.schedule_state_report()

        self._logger.info("function executor controller shutdown finished")
        debug_print_events(events=self._events, logger=self._logger)


def _to_fe_state_metric_label(state: _FE_CONTROLLER_STATE, logger: Any) -> str:
    if state == _FE_CONTROLLER_STATE.NOT_STARTED:
        return METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_NOT_STARTED
    elif state == _FE_CONTROLLER_STATE.STARTING_UP:
        return METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_STARTING_UP
    elif state == _FE_CONTROLLER_STATE.RUNNING:
        return METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_RUNNING
    elif state == _FE_CONTROLLER_STATE.TERMINATING:
        return METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATING
    elif state == _FE_CONTROLLER_STATE.TERMINATED:
        return METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_TERMINATED
    else:
        logger.error(
            "unexpected Function Executor internal state",
            state=state.name,
        )
        return METRIC_FUNCTION_EXECUTORS_WITH_STATE_LABEL_UNKNOWN


_termination_reason_to_short_name_map = {
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_UNKNOWN: "UNKNOWN",
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_INTERNAL_ERROR: "STARTUP_FAILED_INTERNAL_ERROR",
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR: "STARTUP_FAILED_FUNCTION_ERROR",
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_TIMEOUT: "STARTUP_FAILED_FUNCTION_TIMEOUT",
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY: "UNHEALTHY",
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_INTERNAL_ERROR: "INTERNAL_ERROR",
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_TIMEOUT: "FUNCTION_TIMEOUT",
    FunctionExecutorTerminationReason.FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED: "FUNCTION_CANCELLED",
}


def _termination_reason_to_short_name(value: FunctionExecutorTerminationReason) -> str:
    # The enum value names are really long, shorten them to make the logs more readable.
    if value is None:
        return "None"

    return _termination_reason_to_short_name_map.get(value, "UNEXPECTED")


def _to_alloc_result_proto(alloc_info: AllocationInfo, logger: Any) -> AllocationResult:
    allocation: Allocation = alloc_info.allocation
    # Might be None if the alloc wasn't prepared successfully.
    input: AllocationInput | None = alloc_info.input
    # Never None here as we're completing the alloc here.
    output: AllocationOutput | None = alloc_info.output

    execution_duration_ms: int | None = None
    if (
        output.execution_start_time is not None
        and output.execution_end_time is not None
    ):
        # <= 0.99 ms functions get billed as 1 ms.
        execution_duration_ms = math.ceil(
            (output.execution_end_time - output.execution_start_time) * 1000
        )

    request_error_output: DataPayload | None = None
    if output.request_error_output is not None:
        # input can't be None if request_error_output is set because the alloc ran already.
        request_error_output = _so_to_data_payload_proto(
            so=output.request_error_output,
            blob_uri=input.request_error_blob_uri,
            logger=logger,
        )

    output_value: DataPayload | None = None
    if output.output_value is not None:
        # input can't be None if output_value is set because the alloc ran already.
        output_value = _so_to_data_payload_proto(
            so=output.output_value,
            blob_uri=input.function_outputs_blob_uri,
            logger=logger,
        )

    output_updates: ExecutionPlanUpdates | None = None
    if output.output_execution_plan_updates is not None:
        output_updates = _to_execution_plan_updates_proto(
            fe_updates=output.output_execution_plan_updates,
            output_blob_uri=input.function_outputs_blob_uri,
            logger=logger,
        )

    return AllocationResult(
        function=FunctionRef(
            namespace=allocation.function.namespace,
            application_name=allocation.function.application_name,
            function_name=allocation.function.function_name,
            application_version=allocation.function.application_version,
        ),
        allocation_id=allocation.allocation_id,
        function_call_id=allocation.function_call_id,
        request_id=allocation.request_id,
        outcome_code=output.outcome_code,
        failure_reason=output.failure_reason,
        value=output_value,
        updates=output_updates,
        request_error=request_error_output,
        execution_duration_ms=execution_duration_ms,
    )


def _to_execution_plan_updates_proto(
    fe_updates: FEExecutionPlanUpdates, output_blob_uri: str, logger: Any
) -> ExecutionPlanUpdates:
    # TODO: Validate FEExecutionPlanUpdates object.
    executor_updates: List[ExecutionPlanUpdate] = []
    for fe_update in fe_updates.updates:
        fe_update: FEExecutionPlanUpdate

        if fe_update.HasField("function_call"):
            executor_updates.append(
                ExecutionPlanUpdate(
                    function_call=_to_function_call_proto(
                        fe_function_call=fe_update.function_call,
                        output_blob_uri=output_blob_uri,
                        logger=logger,
                    )
                )
            )
        elif fe_update.HasField("reduce"):
            executor_updates.append(
                ExecutionPlanUpdate(
                    reduce=_to_reduce_op_proto(
                        reduce=fe_update.reduce,
                        output_blob_uri=output_blob_uri,
                        logger=logger,
                    )
                )
            )
        else:
            logger.error(
                "unexpected FEExecutionPlanUpdate with no function_call or reduce set",
            )

    return ExecutionPlanUpdates(
        updates=executor_updates,
        root_function_call_id=fe_updates.root_function_call_id,
    )


def _to_function_call_proto(
    fe_function_call: FEFunctionCall, output_blob_uri: str, logger: Any
) -> FunctionCall:
    args: List[FunctionArg] = []
    for fe_arg in fe_function_call.args:
        fe_arg: FEFunctionArg
        args.append(
            _to_function_arg_proto(
                fe_function_arg=fe_arg,
                output_blob_uri=output_blob_uri,
                logger=logger,
            )
        )

    return FunctionCall(
        id=fe_function_call.id,
        target=_to_function_ref_proto(fe_function_call.target),
        args=args,
        call_metadata=fe_function_call.call_metadata,
    )


def _to_reduce_op_proto(
    reduce: FEReduceOp, output_blob_uri: str, logger: Any
) -> ReduceOp:
    collection: List[FunctionArg] = []
    for fe_arg in reduce.collection:
        fe_arg: FEFunctionArg
        collection.append(
            _to_function_arg_proto(
                fe_function_arg=fe_arg,
                output_blob_uri=output_blob_uri,
                logger=logger,
            )
        )

    return ReduceOp(
        id=reduce.id,
        collection=collection,
        reducer=_to_function_ref_proto(reduce.reducer),
        call_metadata=reduce.call_metadata,
    )


def _to_function_ref_proto(fe_function_ref: FEFunctionRef) -> FunctionRef:
    return FunctionRef(
        namespace=fe_function_ref.namespace,
        application_name=fe_function_ref.application_name,
        function_name=fe_function_ref.function_name,
        application_version=fe_function_ref.application_version,
    )


def _to_function_arg_proto(
    fe_function_arg: FEFunctionArg, output_blob_uri: str, logger: Any
) -> FunctionArg:
    if fe_function_arg.HasField("function_call_id"):
        return FunctionArg(function_call_id=fe_function_arg.function_call_id)
    elif fe_function_arg.HasField("value"):
        return FunctionArg(
            inline_data=_so_to_data_payload_proto(
                so=fe_function_arg.value,
                blob_uri=output_blob_uri,
                logger=logger,
            )
        )
    else:
        logger.error(
            "unexpected FEFunctionArg with no value or function_call_id set",
        )
        return FunctionArg()  # Empty arg as we can't raise here.


def _so_to_data_payload_proto(
    so: SerializedObjectInsideBLOB,
    blob_uri: str,
    logger: Any,
) -> DataPayload:
    """Converts a serialized object inside BLOB to into a DataPayload."""
    # TODO: Validate SerializedObjectInsideBLOB.
    return DataPayload(
        uri=blob_uri,
        encoding=_so_to_data_payload_encoding(so.manifest.encoding, logger),
        encoding_version=so.manifest.encoding_version,
        content_type=(
            so.manifest.content_type if so.manifest.HasField("content_type") else None
        ),
        metadata_size=so.manifest.metadata_size,
        offset=so.offset,
        size=so.manifest.size,
        sha256_hash=so.manifest.sha256_hash,
        source_function_call_id=so.manifest.source_function_call_id,
        # id is not used
    )


def _so_to_data_payload_encoding(
    encoding: SerializedObjectEncoding, logger: Any
) -> DataPayloadEncoding:
    if encoding == SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_BINARY_PICKLE:
        return DataPayloadEncoding.DATA_PAYLOAD_ENCODING_BINARY_PICKLE
    elif encoding == SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_UTF8_JSON:
        return DataPayloadEncoding.DATA_PAYLOAD_ENCODING_UTF8_JSON
    elif encoding == SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_UTF8_TEXT:
        return DataPayloadEncoding.DATA_PAYLOAD_ENCODING_UTF8_TEXT
    elif encoding == SerializedObjectEncoding.SERIALIZED_OBJECT_ENCODING_RAW:
        return DataPayloadEncoding.DATA_PAYLOAD_ENCODING_RAW
    else:
        logger.error(
            "unexpected encoding for SerializedObject",
            encoding=SerializedObjectEncoding.Name(encoding),
        )
        return DataPayloadEncoding.DATA_PAYLOAD_ENCODING_UNKNOWN
